# pylint: skip-file
# type: ignore

import re
import sys
from datetime import timedelta, datetime, MINYEAR

import numpy
from PyQt5 import QtCore, QtGui
from PyQt5.QtCore import QTimer, Qt
from PyQt5.QtGui import QPainter, QBrush, QColor
from PyQt5.QtWidgets import QSlider, QDialog, QLineEdit, QPushButton, QHBoxLayout, QLabel, QGridLayout, QWidget

from trnsysGUI.Connection import Connection
import pandas as pd
import itertools
import numpy as np


class MassFlowVisualizer(QDialog):
    def __init__(self, parent, mfrFile, tempFile):

        super(MassFlowVisualizer, self).__init__(parent)

        self.logger = parent.logger

        self.dataFilePath = mfrFile
        self.tempDatafilePath = tempFile
        self.loadedFile = False
        self.tempLoadedFile = False
        self.loadFile()
        self.loadTempFile()
        self.maxTimeStep = len(self.massFlowData.index) - 1
        self.showMass = False

        self.setMinimumSize(1000, 200)

        self.parent = parent
        self.timeStep = 0
        self.timeSteps = len(self.massFlowData.index) - 1  # 0 to 364

        # threshold values for positive list
        self.medianValue = 0
        self.lowerQuarter = 0
        self.upperQuarter = 0
        self.minValue = 0
        self.maxValue = 0

        self.medianValueMfr = 0
        self.lowerQuarterMfr = 0
        self.upperQuarterMfr = 0
        self.minValueMfr = 0
        self.maxValueMfr = 0

        self.getThresholdValues()
        self.getTempThresholdValues()

        self.jumpValue = QLabel("Jump by ( 30 = 1 Hour ):\n              ( 720 = 1 Day )")
        self.jumpValueLE = QLineEdit("1")

        self.slider = QSlider(parent)
        self.setSlider()
        self.slider.sliderReleased.connect(self.testValChange)
        self.slider.sliderPressed.connect(self.pressedSlider)
        self.slider.valueChanged.connect(self.moveValues)
        self.slider.setTickInterval(24)

        self.qtm = QTimer(parent)
        self.lines = None
        self.started = False

        self.paused = True

        nameLabel = QLabel("Name:")
        self.currentStepLabel = QLabel("Time: " + str(self.convertTime(self.getTime(0))))
        self.le = QLineEdit("NONE")

        self.showMassButton = QPushButton("Show mass")  # comment out
        self.togglePauseButton = QPushButton("Pause/Continue")
        self.cancelButton = QPushButton("Cancel")

        minColorLabel = QLabel("Min")
        minColorLabel.setStyleSheet("color: blue")
        maxColorLabel = QLabel("Max")
        maxColorLabel.setStyleSheet("color: red")
        minToFiftyPercentile = QLabel("Min to 50%")
        minToFiftyPercentile.setStyleSheet("color: cyan")
        fiftyPercentileToMax = QLabel("50% to max")
        fiftyPercentileToMax.setStyleSheet("color: pink")
        arrowLabel = QLabel("-->")
        arrowLabel2 = QLabel("-->")
        arrowLabel3 = QLabel("-->")

        colorLayout = QHBoxLayout()
        colorLayout.addWidget(minColorLabel)
        colorLayout.addWidget(arrowLabel)
        colorLayout.addWidget(minToFiftyPercentile)
        colorLayout.addWidget(arrowLabel2)
        colorLayout.addWidget(fiftyPercentileToMax)
        colorLayout.addWidget(arrowLabel3)
        colorLayout.addWidget(maxColorLabel)
        colorLayout.addStretch()

        buttonLayout = QHBoxLayout()
        buttonLayout.addStretch()
        buttonLayout.addWidget(self.jumpValue)
        buttonLayout.addWidget(self.jumpValueLE)
        buttonLayout.addWidget(self.showMassButton)  # comment out
        buttonLayout.addWidget(self.togglePauseButton)
        buttonLayout.addWidget(self.cancelButton)
        layout = QGridLayout()
        layout.addWidget(nameLabel, 0, 0)
        layout.addWidget(self.le, 0, 1)
        layout.addLayout(colorLayout, 1, 0, 1, 0)
        layout.addLayout(buttonLayout, 2, 0, 2, 0)
        layout.addWidget(self.currentStepLabel, 3, 0, 1, 2)
        layout.addWidget(self.slider, 4, 0, 1, 2)  # Only for debug (Why do I need a 3 here instead of a 2 for int:row?)

        self.setLayout(layout)

        self.showMassButton.clicked.connect(self.showMassBtn)  # comment out
        self.togglePauseButton.clicked.connect(self.togglePause)
        self.cancelButton.clicked.connect(self.cancel)

        self.advance()

        self.setWindowTitle("Flow visualizer")
        ph = parent.geometry().height()
        pw = parent.geometry().width()
        px = parent.geometry().x()
        py = parent.geometry().y()
        dw = self.width()
        dh = self.height()
        self.move(parent.centralWidget.diagramView.geometry().topLeft())
        self.show()

    def togglePause(self):
        if self.paused:
            self.continueVis()
        else:
            self.pauseVis()

    def cancel(self):
        self.pauseVis()
        self.close()
        self.parent.centralWidget.updateConnGrads()

    # comment out
    def showMassBtn(self):
        """
        For showing the mass under the connections
        -------

        """
        self.showMass = not self.showMass

        for t in self.parent.centralWidget.trnsysObj:
            if isinstance(t, Connection):
                if self.showMass:
                    t.firstS.labelMass.setVisible(True)
                else:
                    t.firstS.labelMass.setVisible(False)

        self.logger.debug("%s %s %s" % (str(self.minValue), str(self.medianValue), str(self.maxValue)))

    def loadFile(self):
        if not self.loadedFile:
            self.massFlowData = pd.read_csv(self.dataFilePath, sep="\t").rename(columns=lambda x: x.strip())
        self.loadedFile = True

    def loadTempFile(self):
        if not self.tempLoadedFile:
            self.tempMassFlowData = pd.read_csv(self.tempDatafilePath, sep="\t").rename(columns=lambda x: x.strip())
        self.tempLoadedFile = True

    def start(self):

        self.paused = False
        self.qtm = QTimer(self.parent)
        self.qtm.timeout.connect(self.advance)
        self.qtm.timeout.connect(self.increaseValue)
        self.qtm.start(1000)

    def advance(self):
        if self.timeStep == self.maxTimeStep:
            self.logger.debug("reached end of data, returning")
            self.qtm.stop()
            return

        if self.loadedFile:
            i = 0
            for t in self.parent.centralWidget.trnsysObj:
                if isinstance(t, Connection):
                    if (
                        "Mfr" + t.displayName in self.massFlowData.columns
                        and "T" + t.displayName in self.tempMassFlowData
                    ):
                        mass = str(round(self.massFlowData["Mfr" + t.displayName].iloc[self.timeStep]))
                        mass1Dp = str(round(self.massFlowData["Mfr" + t.displayName].iloc[self.timeStep], 1))
                        temperature = str(round(self.tempMassFlowData["T" + t.displayName].iloc[self.timeStep]))
                        self.logger.debug("Found connection in ts " + str(self.timeStep) + " " + str(i))
                        self.logger.debug("mass flow value of %s : " % t.displayName)
                        t.setMassAndTemperature(mass1Dp, temperature)
                        thickValue = self.getThickness(mass)
                        self.logger.debug("Thickvalue: " + str(thickValue))
                        if self.massFlowData["Mfr" + t.displayName].iloc[self.timeStep] == 0:
                            t.setColor(thickValue, mfr="ZeroMfr")
                        elif round(abs(self.tempMassFlowData["T" + t.displayName].iloc[self.timeStep])) == self.maxValue:
                            t.setColor(thickValue, mfr="max")
                        elif round(abs(self.tempMassFlowData["T" + t.displayName].iloc[self.timeStep])) == self.minValue:
                            t.setColor(thickValue, mfr="min")
                        elif (
                            self.minValue
                            < round(abs(self.tempMassFlowData["T" + t.displayName].iloc[self.timeStep]))
                            <= self.lowerQuarter
                        ):
                            t.setColor(thickValue, mfr="minTo25")
                        elif (
                            self.lowerQuarter
                            < round(abs(self.tempMassFlowData["T" + t.displayName].iloc[self.timeStep]))
                            <= self.medianValue
                        ):
                            t.setColor(thickValue, mfr="25To50")
                        elif (
                            self.medianValue
                            < round(abs(self.tempMassFlowData["T" + t.displayName].iloc[self.timeStep]))
                            <= self.upperQuarter
                        ):
                            t.setColor(thickValue, mfr="50To75")
                        elif (
                            self.upperQuarter
                            < round(abs(self.tempMassFlowData["T" + t.displayName].iloc[self.timeStep]))
                            < self.maxValue
                        ):
                            t.setColor(thickValue, mfr="75ToMax")
                        else:
                            t.setColor(thickValue, mfr="test")
                        i += 1
                else:
                    if "xFrac" + t.displayName in self.massFlowData.columns:
                        valvePosition = str(self.massFlowData["xFrac" + t.displayName].iloc[self.timeStep])
                        t.setPositionForMassFlowSolver(valvePosition)
                        t.posLabel.setPlainText(valvePosition)
                        self.logger.debug("valve position: " + str(valvePosition))

        else:
            return

    def pauseVis(self):
        self.paused = True
        self.qtm.stop()

    def continueVis(self):
        if self.started:
            self.paused = False
            self.qtm.start(1000)
        else:
            self.start()

    def setSlider(self):
        self.slider.setOrientation(1)
        self.slider.setMinimum(0)
        self.slider.setMaximum(self.timeSteps)
        self.slider.setTickInterval(1)
        self.slider.setTickPosition(2)
        if self.checkTimeStep and self.checkTempTimeStep():
            self.slider.setVisible(True)
            self.slider.setEnabled(False)
            self.increaseValue()
        else:
            self.slider.setVisible(True)
            self.slider.setEnabled(True)

    def testValChange(self):
        """
        Updates mass flow visualizer after releasing the mouse
        """
        val = self.slider.value()
        self.logger.debug("Slider value has changed to " + str(val))
        time = self.getTime(val)
        self.currentStepLabel.setText("Time :" + str(self.convertTime(time)))
        self.timeStep = val
        self.advance()

    def moveValues(self):
        """
        Updates mass flow visualizer when moving the mouse
        """
        val = self.slider.value()
        self.logger.debug("Slider value is still: " + str(val))
        time = self.getTime(val)
        self.currentStepLabel.setText("Time :" + str(self.convertTime(time)))
        self.timeStep = val
        self.advance()

    def increaseValue(self):
        """
        For automatic slider movement

        """

        self.timeStep += float(self.jumpValueLE.text())
        if self.timeStep > self.maxTimeStep:
            self.timeStep = 0
        self.slider.setValue(self.timeStep)

    def decreaseValue(self):
        """
        For automatic slider movement

        """

        self.timeStep -= float(self.jumpValueLE.text())
        if self.timeStep < 0:
            self.timeStep = self.maxTimeStep
        self.slider.setValue(self.timeStep)

    def checkTimeStep(self):
        """
        Check individual columns of the data frame, If a column has rows with different values, return False.
        Else, continue to next column. Return True if no such column can be found.

        False indicates the rows are not identical and slider should be enabled
        True indicates the rows are identical and slider should be disabled

        """
        massFlowDataDup = self.massFlowData
        massFlowDataDup = massFlowDataDup.drop(massFlowDataDup.index[0])
        for items in round(abs(massFlowDataDup)).nunique().iteritems():
            if items[0] != "TIME":
                if items[1] > 1:
                    return False
        return True

    def checkTempTimeStep(self):
        tempMassFlowDataDup = self.tempMassFlowData
        tempMassFlowDataDup = tempMassFlowDataDup.drop(tempMassFlowDataDup.index[0])
        for items in tempMassFlowDataDup.nunique().iteritems():
            if items[0] != "TIME":
                if items[1] > 1:
                    return False
        return True

    def getThresholdValues(self):
        """
        Access the data frame, convert into a nested list.
        Merge the nested list into one single list.
        Remove 'nan' from the list.
        Convert the values inside the list into absolute values then round them off.
        Split the list into two for negative and positive values.
        Get the minimum value, the 25th percentile, the median value, the 75th percentile and
        the max value from the lists.
        Returns
        -------
        """

        data = self.massFlowData.values.tolist()  # data frame converted to nested list
        for sublist in data:  # delete the time column from the list
            del sublist[0]
        data = list(itertools.chain.from_iterable(data))  # nested list combined into one list
        cleanedData = [x for x in data if str(x) != "nan"]  # remove nan from list
        cleanedData = [round(abs(num)) for num in cleanedData]  # get absolute value and round off
        nonZeroData = [x for x in cleanedData if x > 1]  # a work around to remove the 1 values from the data frame
        noDuplicateData = list(dict.fromkeys(nonZeroData))

        self.medianValueMfr = np.percentile(noDuplicateData, 50)  # median value / 50th percentile
        self.lowerQuarterMfr = np.percentile(noDuplicateData, 25)  # 25th percentile
        self.upperQuarterMfr = np.percentile(noDuplicateData, 75)  # 75th percentile
        self.minValueMfr = np.min(noDuplicateData)  # minimum value excluding 0
        self.maxValueMfr = np.max(noDuplicateData)  # max value

    def getThickness(self, mass):
        mass = abs(float(mass))
        if mass == self.minValueMfr:
            return 2
        elif self.minValueMfr < mass <= self.medianValueMfr:
            return 4
        elif self.medianValueMfr < mass < self.maxValueMfr:
            return 6
        elif mass == self.maxValueMfr:
            return 8
        else:
            return 2

    def getTempThresholdValues(self):
        """
        Access the data frame, convert into a nested list.
        Merge the nested list into one single list.
        Remove 'nan' from the list.
        Convert the values inside the list into absolute values then round them off.
        Split the list into two for negative and positive values.
        Get the minimum value, the 25th percentile, the median value, the 75th percentile and
        the max value from the lists.
        Returns
        -------
        """

        data = self.tempMassFlowData.values.tolist()  # data frame converted to nested list
        for sublist in data:  # delete the time column from the list
            del sublist[0]
        data = list(itertools.chain.from_iterable(data))  # nested list combined into one list
        cleanedData = [x for x in data if str(x) != "nan"]  # remove nan from list
        cleanedData = [round(abs(num)) for num in cleanedData]  # get absolute value and round off
        noDuplicateData = list(dict.fromkeys(cleanedData))

        self.medianValue = np.percentile(noDuplicateData, 50)  # median value / 50th percentile
        self.lowerQuarter = np.percentile(noDuplicateData, 25)  # 25th percentile
        self.upperQuarter = np.percentile(noDuplicateData, 75)  # 75th percentile
        self.minValue = np.min(noDuplicateData)  # minimum value excluding 0
        self.maxValue = np.max(noDuplicateData)  # max value

        # print(noDuplicateData)
        # sys.exit()

    def getTime(self, row):
        """
        Gets the time of the current time step
        """
        data = self.massFlowData
        timeColumn = data[data.columns[0]]
        self.logger.debug(str(timeColumn[row]))
        return timeColumn[row]

    def convertTime(self, time):
        """
        Convert the time into YYYY--MM--DD HH:MM:SS format
        """
        noOfHours = 8760
        decHour = float(time) / float(noOfHours)
        base = datetime(MINYEAR, 1, 1)
        result = base + timedelta(seconds=(base.replace(year=base.year + 1) - base).total_seconds() * decHour)
        return str(result)

    def pressedSlider(self):
        self.pauseVis()

    def closeEvent(self, a0):
        for t in self.parent.centralWidget.trnsysObj:
            if isinstance(t, Connection):
                t.firstS.labelMass.setVisible(False)

        self.pauseVis()
        self.parent.centralWidget.updateConnGrads()
        self.parent.massFlowEnabled = False

        super(MassFlowVisualizer, self).closeEvent(a0)

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Up:
            self.logger.debug("Up is pressed")
            self.increaseValue()
        elif e.key() == Qt.Key_Down:
            self.logger.debug("Down is pressed")
            self.decreaseValue()



# pylint: skip-file
# type: ignore

from PyQt5.QtWidgets import QDialog, QDialogButtonBox, QVBoxLayout, QLabel


class TestDlg(QDialog):
    """
    This is the dialog box for testing. It will be called when an exported file cannot be found inside the reference
    folder. (file of same name doesnt exist inside reference folder)
    """

    def __init__(self, exportedFile, *args, **kwargs):
        super(TestDlg, self).__init__(*args, **kwargs)

        self.exportBool = False

        self.setWindowTitle("Error!")
        self.dlgMessage = QLabel()
        self.strMessage = "%s not found inside Reference folder, add the file into Reference folder?" % exportedFile
        self.dlgMessage.setText(self.strMessage)

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.acceptExport)
        self.buttonBox.rejected.connect(self.reject)

        self.layout = QVBoxLayout()
        self.layout.addWidget(self.dlgMessage)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)
        self.exec_()

    def acceptExport(self):
        self.exportBool = True
        self.accept()



# pylint: skip-file
# type: ignore

from PyQt5.QtWidgets import QDialog, QLabel, QLineEdit, QPushButton, QHBoxLayout, QGridLayout


class diagramDlg(QDialog):
    def __init__(self, parent):
        super(diagramDlg, self).__init__(parent)
        self.diag = parent
        nameLabel = QLabel("Diagram name:")
        objectLabel = QLabel("Object:" + str(self.diag))
        self.le = QLineEdit(self.diag.diagramName)

        self.okButton = QPushButton("OK")
        self.cancelButton = QPushButton("Cancel")

        buttonLayout = QHBoxLayout()
        buttonLayout.addStretch()
        buttonLayout.addWidget(self.okButton)
        buttonLayout.addWidget(self.cancelButton)
        layout = QGridLayout()
        layout.addWidget(nameLabel, 0, 0)
        layout.addWidget(self.le, 0, 1)
        layout.addLayout(buttonLayout, 1, 0, 2, 0)
        # layout.addWidget(objectLabel, 3, 0, 1, 2)  # Only for debug (Why do I need a 3 here instead of a 2 for int:row?)
        self.setLayout(layout)

        self.okButton.clicked.connect(self.acceptedEdit)
        self.cancelButton.clicked.connect(self.cancel)
        self.setWindowTitle("Diagram Properties")
        self.show()

    def acceptedEdit(self):
        self.diag.renameDiagram(self.le.text())
        self.close()

    def cancel(self):
        self.close()



# pylint: skip-file
# type: ignore

import setuptools
import pathlib as _pl
import itertools as _it
import dataclasses as _dc

with open("README.md", "r") as fh:
    long_description = fh.read()


@_dc.dataclass
class _DestDirSourceFilePath:
    destDir: str
    sourceFilePath: str


def _getDataFilePairs():
    dataDirPath = _pl.Path(__file__).parent / "data"

    dataFilePaths = [
        p.relative_to(dataDirPath) for p in dataDirPath.rglob("*") if p.is_file()
    ]

    destDirSourcePathPairs = [
        _DestDirSourceFilePath(str("pytrnsys_gui_data" / p.parent), str("data" / p)) for p in dataFilePaths
    ]

    sortedPairs = sorted(destDirSourcePathPairs, key=lambda dp: dp.destDir)

    dataFilePairs = [
        (d, [dp.sourceFilePath for dp in dps])
        for d, dps in _it.groupby(sortedPairs, key=lambda dp: dp.destDir)
    ]

    return dataFilePairs


setuptools.setup(
    name="pytrnsys-gui",
    author="Martin Neugebauer",
    version_config=True,
    author_email="martin.neugebauer@ost.ch",
    description="A GUI for Trnsys",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SPF-OST/pytrnsys",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={
        "trnsysGUI": [
            "images/*.png",
            "images/*.svg",
            "templates/run.config",
            "templates/generic/*.ddck",
        ]
    },
    data_files=_getDataFilePairs(),
    entry_points={"console_scripts": ["pytrnsys-gui=trnsysGUI.GUI:main"]},
    install_requires=[
        "pytrnsys",
        "PyQT5",
        "matplotlib",
        "numpy",
        "pandas",
        "bokeh",
        "dataclasses_jsonschema",
        "appdirs",
    ],
    setup_requires=["setuptools-git-versioning"],
    python_requires=">=3.9",
)



import dataclasses as _dc
import uuid as _uuid
import typing as _tp
import enum as _enum

import dataclasses_jsonschema as _dcj

import trnsysGUI.serialization as _ser
import trnsysGUI.IdGenerator as _id


@_dc.dataclass
class StorageTankVersion0(_ser.UpgradableJsonSchemaMixinVersion0):
    FlippedH: bool
    FlippedV: bool
    BlockName: str
    BlockDisplayName: str
    size_h: int
    StoragePosition: _tp.Tuple[float, float]
    trnsysID: int
    ID: int
    GroupName: str
    HxList: _tp.Sequence["HeatExchangerLegacyVersion"]
    PortPairList: _tp.Sequence["DirectPortPairLegacyVersion"]

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("05f422d3-41fd-48d1-b8d0-4655d9f65247")


@_dc.dataclass
class HeatExchangerLegacyVersion(_dcj.JsonSchemaMixin):
    SideNr: int
    Width: float
    Height: float
    Offset: _tp.Tuple[float, float]
    DisplayName: str
    ParentID: int
    Port1ID: int
    Port2ID: int
    connTrnsysID: int
    ID: int


@_dc.dataclass
class DirectPortPairLegacyVersion(_dcj.JsonSchemaMixin):
    Side: bool
    Port1offset: float
    Port2offset: float
    Port1ID: int
    Port2ID: int
    ConnID: int
    ConnCID: int
    ConnDisName: str
    trnsysID: int


@_dc.dataclass
class StorageTank(_ser.UpgradableJsonSchemaMixin):
    isHorizontallyFlipped: bool
    isVerticallyFlipped: bool

    BlockName: str
    BlockDisplayName: str

    groupName: str
    id: int
    trnsysId: int

    height: int
    position: _tp.Tuple[float, float]

    heatExchangers: _tp.Sequence["HeatExchanger"]

    directPortPairs: _tp.Sequence["DirectPortPair"]

    @classmethod
    def from_dict(
        cls,
        data: _dcj.JsonDict,
        validate=True,
        validate_enums: bool = True,
    ) -> "StorageTank":
        data.pop(".__BlockDict__")
        storageTank = super().from_dict(data, validate, validate_enums)
        return _tp.cast(StorageTank, storageTank)

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,  # pylint: disable=duplicate-code
    ) -> _dcj.JsonDict:
        data = super().to_dict(omit_none, validate, validate_enums)
        data[".__BlockDict__"] = True
        return data

    @classmethod
    def upgrade(
        cls, superseded: _ser.UpgradableJsonSchemaMixinVersion0
    ) -> "StorageTank":
        if not isinstance(superseded, StorageTankVersion0):
            raise ValueError(
                f"Can only upgrade a {StorageTankVersion0.__name__} instance."
            )

        heatExchangers = cls._upgradeHeatExchangers(
            superseded.HxList, superseded.size_h
        )
        directPortPairs = cls._upgradeDirectPortPairs(
            superseded.PortPairList, superseded.size_h
        )

        return StorageTank(
            superseded.FlippedH,
            superseded.FlippedV,
            superseded.BlockName,
            superseded.BlockDisplayName,
            superseded.GroupName,
            superseded.ID,
            superseded.trnsysID,
            superseded.size_h,
            superseded.StoragePosition,
            heatExchangers,
            directPortPairs,
        )

    @classmethod
    def _upgradeHeatExchangers(
        cls,
        supersededHeatExchangers: _tp.Sequence[HeatExchangerLegacyVersion],
        storageTankHeight: float,
    ) -> _tp.Sequence["HeatExchanger"]:
        heatExchangers = []
        for supersededHeatExchanger in supersededHeatExchangers:
            heatExchangerLegacyVersion = HeatExchangerVersion0.createFromLegacyHeatExchanger(
                supersededHeatExchanger, storageTankHeight
            )

            heatExchanger = HeatExchanger.fromInstance(heatExchangerLegacyVersion)

            heatExchangers.append(heatExchanger)

        return heatExchangers

    @classmethod
    def _upgradeDirectPortPairs(
        cls,
        supersededPairs: _tp.Sequence[DirectPortPairLegacyVersion],
        storageTankHeight: float,
    ) -> _tp.Sequence["DirectPortPair"]:
        directPortPairs = []
        for supersededPair in supersededPairs:
            directPortPairLegacyVersion = DirectPortPairVersion0.createFromLegacyPortPair(
                    supersededPair, storageTankHeight
            )

            directPortPair = DirectPortPair.fromInstance(directPortPairLegacyVersion)

            directPortPairs.append(directPortPair)

        return directPortPairs

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("05f422d3-41fd-48d1-b8d0-4655d9f65247")

    @classmethod
    def getSupersededClass(cls):
        return StorageTankVersion0


class Side(_enum.Enum):
    LEFT = "left"
    RIGHT = "right"

    @staticmethod
    def createFromSideNr(sideNr: int) -> "Side":
        if sideNr == 2:
            return Side.RIGHT

        if sideNr == 0:
            return Side.LEFT

        raise ValueError(f"Unknown side number: {sideNr}")

    def toSideNr(self):
        if self == self.LEFT:
            return 0

        if self == self.RIGHT:
            return 2

        raise ValueError("Cannot convert {self} to side nr.")


@_dc.dataclass
class Port(_dcj.JsonSchemaMixin):
    id: int
    relativeHeight: float


@_dc.dataclass
class PortPairVersion0(_ser.UpgradableJsonSchemaMixinVersion0):
    side: Side

    name: str

    inputPort: Port
    outputPort: Port

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID('e1b81a7b-51ca-44f0-9d3a-6b6c42eb8e91')


@_dc.dataclass
class PortPair(_ser.UpgradableJsonSchemaMixin):
    side: Side

    trnsysId: int
    inputPort: Port
    outputPort: Port

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixinVersion0]:
        return PortPairVersion0

    @classmethod
    def upgrade(cls, superseded: _ser.UpgradableJsonSchemaMixinVersion0) -> "PortPair":
        if not isinstance(superseded, PortPairVersion0):
            raise ValueError(
                f"Superseded instance is not of type {PortPairVersion0.__name__}"
            )

        return PortPair(
            superseded.side,
            _id.IdGenerator.UNINITIALIZED_ID,  # type: ignore[attr-defined]
            superseded.inputPort,
            superseded.outputPort
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID('ed17e7a7-4798-4a80-beb7-81730523cf35')


@_dc.dataclass
class HeatExchangerVersion0(_ser.UpgradableJsonSchemaMixinVersion0):
    portPair: PortPairVersion0

    width: float

    parentId: int

    id: int
    connectionTrnsysId: int

    @classmethod
    def createFromLegacyHeatExchanger(
        cls,
        superseded: HeatExchangerLegacyVersion,
        storageTankHeight: float,
    ) -> "HeatExchangerVersion0":
        absoluteInputHeight = storageTankHeight - superseded.Offset[1]
        absoluteOutputHeight = absoluteInputHeight - superseded.Height

        relativeInputHeight = round(absoluteInputHeight / storageTankHeight, 2)
        relativeOutputHeight = round(absoluteOutputHeight / storageTankHeight, 2)

        inputPort = Port(superseded.Port1ID, relativeInputHeight)
        outputPort = Port(superseded.Port2ID, relativeOutputHeight)

        side = Side.createFromSideNr(superseded.SideNr)

        portPair = PortPairVersion0(side, superseded.DisplayName, inputPort, outputPort)

        heatExchanger = HeatExchangerVersion0(
            portPair,
            superseded.Width,
            superseded.ParentID,
            superseded.ID,
            superseded.connTrnsysID,
        )

        return heatExchanger

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("23938d79-dc02-4752-ba1b-25d7a610de27")


@_dc.dataclass
class HeatExchangerVersion1(_ser.UpgradableJsonSchemaMixin):
    portPair: PortPairVersion0
    width: float
    parentId: int
    id: int

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixinVersion0]:
        return HeatExchangerVersion0

    @classmethod
    def upgrade(cls, superseded: _ser.UpgradableJsonSchemaMixinVersion0) -> "HeatExchangerVersion1":
        if not isinstance(superseded, HeatExchangerVersion0):
            raise ValueError(
                f"`superseded` is not of type {HeatExchangerVersion0.__name__}"
            )

        return HeatExchangerVersion1(
            superseded.portPair, superseded.width, superseded.parentId, superseded.id
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("68c5cebb-0c47-4dec-8c85-8872b7f6c238")


@_dc.dataclass
class HeatExchanger(_ser.UpgradableJsonSchemaMixin):
    portPair: PortPair
    name: str
    width: float
    parentId: int
    id: int

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixinVersion0]:
        return HeatExchangerVersion1

    @classmethod
    def upgrade(cls, superseded: _ser.UpgradableJsonSchemaMixinVersion0) -> "HeatExchanger":
        if not isinstance(superseded, HeatExchangerVersion1):
            raise ValueError(
                f"`superseded` is not of type {HeatExchangerVersion1.__name__}"
            )

        portPair = PortPair.fromInstance(superseded.portPair)

        return HeatExchanger(
            portPair, superseded.portPair.name, superseded.width, superseded.parentId, superseded.id
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID('be068f84-2627-4f53-87a9-9a1f4fd1c529')


@_dc.dataclass
class DirectPortPairVersion0(_ser.UpgradableJsonSchemaMixinVersion0):
    portPair: PortPairVersion0

    id: int

    connectionId: int
    trnsysId: int

    @classmethod
    def createFromLegacyPortPair(
        cls, superseded: DirectPortPairLegacyVersion, storageTankHeight: float
    ) -> "DirectPortPairVersion0":
        absoluteInputHeight = storageTankHeight - superseded.Port1offset
        relativeInputHeight = round(absoluteInputHeight / storageTankHeight, 2)
        inputPort = Port(superseded.Port1ID, relativeInputHeight)

        absoluteOutputHeight = storageTankHeight - superseded.Port2offset
        relativeOutputHeight = round(absoluteOutputHeight / storageTankHeight, 2)
        outputPort = Port(superseded.Port2ID, relativeOutputHeight)

        side = Side.LEFT if superseded.Side else Side.RIGHT
        portPair = PortPairVersion0(side, superseded.ConnDisName, inputPort, outputPort)

        return DirectPortPairVersion0(
            portPair,
            superseded.ConnID,
            superseded.ConnCID,
            superseded.trnsysID,
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("2629b768-3cfe-4bff-b28d-658ab6154202")


@_dc.dataclass
class DirectPortPair(_ser.UpgradableJsonSchemaMixin):
    portPair: PortPair

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[_ser.UpgradableJsonSchemaMixinVersion0]:
        return DirectPortPairVersion0

    @classmethod
    def upgrade(cls, superseded: _ser.UpgradableJsonSchemaMixinVersion0) -> "DirectPortPair":
        if not isinstance(superseded, DirectPortPairVersion0):
            raise ValueError(
                f"Superseded instance is not of type {DirectPortPairVersion0.__name__}"
            )

        portPair = PortPair.fromInstance(superseded.portPair)

        return DirectPortPair(portPair)

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("67aaee05-7d09-40e1-a6e2-d367c7196832")



# pylint: skip-file
# type: ignore

import os
import numpy as num


class Type1924_TesPlugFlow:
    def __init__(self):

        self.listParHx = ["", "", "", ""]
        self.listParPort = num.zeros(10)
        self.listParUa = num.zeros(10)

        self.nMaxPorts = 10
        self.nMaxHx = 6
        self.nMaxSensor = 10
        self.nMaxAvgSensor = 5

        self.sLine = "*************************************\n"
        self.extension = "ddck"

        self.setDefault()

    def setInputs(self, inputs, connectorsPort, connectorsHx, connectorsAux):

        self.inputs = inputs
        self.connectorsPort = connectorsPort
        self.connectorsHx = connectorsHx
        self.connectorsAux = connectorsAux

    def setDefault(self):
        self.Ua = [1, 1, 1, 1, 1]

    def getOneHxInputs(self, nTes, idHx, connectorHx):

        lines = ""

        line = "EQUATIONS 3\n"
        lines = lines + line
        line = "Thx%dIn_Tes%d = %s ! @connectDdck\n" % (idHx, nTes, connectorHx[idHx - 1]["T"])
        lines = lines + line
        line = "Mfrhx%d_Tes%d = %s ! @connectDdck\n" % (idHx, nTes, connectorHx[idHx - 1]["Mfr"])
        lines = lines + line
        line = "Thx%dInRev_Tes%d = %s ! @connectDdck\n" % (idHx, nTes, connectorHx[idHx - 1]["Trev"])
        lines = lines + line

        return lines

    @staticmethod
    def _getOnePortInputs(nTes, idPort, connectorPort):

        lines = ""

        line = "EQUATIONS 3\n"
        lines = lines + line
        line = "Tdp%dIn_Tes%d = %s ! @connectDdck\n" % (idPort, nTes, connectorPort[idPort - 1]["T"])
        lines = lines + line
        line = "Mfrdp%d_Tes%d = %s ! @connectDdck\n" % (idPort, nTes, connectorPort[idPort - 1]["Mfr"])
        lines = lines + line
        line = "Tdp%dInRev_Tes%d = %s ! @connectDdck\n" % (idPort, nTes, connectorPort[idPort - 1]["Trev"])
        lines = lines + line

        return lines

    def getOnePortParConn(self, idPort, nTes, connectorPort):

        lines = "*********Connecting values of DIRECT PORT=%d***********\n" % (idPort + 1)

        line = "CONSTANTS 2\n"
        lines = lines + line

        line = "zInDp%d_Tes%d=%.2f ! @connectDdck \n" % (idPort + 1, nTes, connectorPort[idPort]["zIn"])
        lines = lines + line
        line = "zOutDp%d_Tes%d=%.2f ! @connectDdck \n" % (idPort + 1, nTes, connectorPort[idPort]["zOut"])
        lines = lines + line

        return lines

    def getOnePortPar(self, idPort, nTes):

        lines = "*********Constant of DIRECT PORT=%d***********\n" % (idPort + 1)

        line = "CONSTANTS 1\n"
        lines = lines + line
        line = "Dp%dStrat_Tes%d=0 ! 0: no forced stratification ; 1: force to stratify\n" % (idPort + 1, nTes)
        lines = lines + line

        return lines

    def getSensorPositionValues(self, nTes):

        lines = self.sLine + "** USER DEFINED TEMPERATURE SENSOR HEIGHTS. To be changed by user \n" + self.sLine

        line = "CONSTANTS 10\n"
        lines = lines + line

        z = 0.05
        for i in range(10):
            line = "zSen%d_Tes%d=%.2f\n" % (i + 1, nTes, z)
            lines = lines + line
            z = z + 0.1

        return lines

    def getOneHxParConnValues(self, nTes, idHx, connectHx):

        lines = "*********Connecting values of HX=%d***********\n" % idHx

        line = "CONSTANTS 4\n"
        lines = lines + line

        line = "zInhx%d_Tes%d=%.2f  ! @connectDdck\n" % (idHx, nTes, connectHx[idHx - 1]["zIn"])
        lines = lines + line
        line = "zOuthx%d_Tes%d=%.2f ! @connectDdck\n" % (idHx, nTes, connectHx[idHx - 1]["zOut"])
        lines = lines + line
        line = "Cphx%d_Tes%d=%s     ! @connectDdck\n" % (idHx, nTes, connectHx[idHx - 1]["cp"])
        lines = lines + line
        line = "Rhohx%d_Tes%d=%s    ! @connectDdck\n" % (idHx, nTes, connectHx[idHx - 1]["rho"])
        lines = lines + line

        return lines

    def getOneHxParValues(self, nTes, idHx):

        lines = "*********Constant values of HX=%d***********\n" % idHx

        line = "CONSTANTS 15\n"
        lines = lines + line

        line = "dInHx%d_Tes%d=0.05 ! m only if modHx%d=1\n" % (idHx, nTes, idHx)
        lines = lines + line
        line = "dOutHx%d_Tes%d=0.06 ! m only if modHx%d=1\n" % (idHx, nTes, idHx)
        lines = lines + line

        line = "LHx%d_Tes%d=20 ! m only if modHx%d=1\n" % (idHx, nTes, idHx)
        lines = lines + line
        line = "LamHx%d_Tes%d=50 ! W/mK only if modHx%d=1\n" % (idHx, nTes, idHx)
        lines = lines + line
        line = "brineConcHx%d_Tes%d=30 ! [0-100] only if modHx%d=1\n" % (idHx, nTes, idHx)
        lines = lines + line

        line = "VHx%d_Tes%d=0. \n" % (idHx, nTes)
        lines = lines + line
        line = "nCvHx%d_Tes%d=20 \n" % (idHx, nTes)
        lines = lines + line

        line = "modHx%d_Tes%d=1 ! modHX1, 0 = physical model, 1 = Drueck-model (Multiport)\n" % (idHx, nTes)
        lines = lines + line
        line = "nNuHx%d_Tes%d=0.5 !  only if modHx%d=1\n" % (idHx, nTes, idHx)
        lines = lines + line
        line = "cNuHx%d_Tes%d=0.25 !  only if modHx%d=1\n" % (idHx, nTes, idHx)
        lines = lines + line

        line = "dUaMfrHx%d_Tes%d=0.375 ! only if modHx%d=0\n" % (idHx, nTes, idHx)
        lines = lines + line
        line = "dUadTHx%d_Tes%d=0.0 ! only if modHx%d=0\n" % (idHx, nTes, idHx)
        lines = lines + line
        line = "dUaTHx%d_Tes%d=0.458 ! only if modHx%d=0\n" % (idHx, nTes, idHx)
        lines = lines + line
        line = "UaHx%d_Tes%d=1105*ratioTes%d ! kJ/hK only if modHx%d=0\n" % (idHx, nTes, nTes, idHx)
        lines = lines + line

        line = "startUpHx%d_Tes%d=0. ! only if modHx%d=0\n" % (idHx, nTes, idHx)
        lines = lines + line

        return lines

    def getHxParValues(self, nTes, nHx):

        lines = ""
        for i in range(self.nMaxHx):
            idHx = i + 1
            line = "** Parameters for heat Exchanger number %d\n" % idHx
            lines = lines + line
            if idHx <= nHx:
                line = "zInHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "zOutHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "dInHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "dOutHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "LHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "LamHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "brineConcHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "VHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "CpHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "RhoHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "nCvHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "modHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "nNuHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "cNuHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "dUaMfrHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "dUadTHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "dUaTHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "UaHx%d_Tes%d " % (idHx, nTes)
                lines = lines + line
                line = "startUpHx%d_Tes%d ! Heax exchanger %d\n" % (idHx, nTes, idHx)
                lines = lines + line
            else:
                line = "-1 -1 "
                lines = lines + line  # zin and zout = -1 when not used
                for i in range(17):
                    line = "zero "
                    lines = lines + line
                lines = lines + "! Heax exchanger %d\n" % (idHx)

        return lines

    def getUaPar(self, nTes):

        lines = ""

        line = "UaBot_Tes%d ! W/k \n" % (nTes)
        lines = lines + line
        line = "Uaz1_Tes%d  ! W/k\n" % (nTes)
        lines = lines + line
        line = "Uaz2_Tes%d  ! W/k\n" % (nTes)
        lines = lines + line
        line = "Uaz3_Tes%d  ! W/k\n" % (nTes)
        lines = lines + line
        line = "UaTop_Tes%d ! W/k\n" % (nTes)
        lines = lines + line

        return lines

    def getUaParValues(self, nTes, Ua):

        lines = ""
        myList = []
        line = "CONSTANTS 10\n"
        lines = lines + line

        line = "Ufoam_Tes%d= 0.67 ! W/(m2K) 6 cm of foam of 0.04 W/(mK) \n" % (nTes)
        lines = lines + line
        line = "Ubot_Tes%d = 1.5 ! W/(m2K) 2 cm of foam of 0.04 W/(mK)\n" % (nTes)
        lines = lines + line
        line = "Atop_Tes%d = Vol_Tes%d/Heigh_Tes%d ! m2\n" % (nTes, nTes, nTes)
        lines = lines + line
        line = "Diameter_Tes%d = (4*ATop_Tes%d/PI)^0.5 ! m \n" % (nTes, nTes)
        lines = lines + line
        line = "ALat_Tes%d = Heigh_Tes%d*PI*Diameter_Tes%d ! m2\n" % (nTes, nTes, nTes)
        lines = lines + line

        line = "UaBot_Tes%d= Ubot_Tes%d*ATop_Tes%d ! @userDefined W/k \n" % (nTes, nTes, nTes)
        lines = lines + line
        line = "Uaz1_Tes%d = Ufoam_Tes%d*ALat_Tes%d/3 ! @userDefined W/k\n" % (nTes, nTes, nTes)
        lines = lines + line
        line = "Uaz2_Tes%d = Ufoam_Tes%d*ALat_Tes%d/3 ! @userDefined W/k\n" % (nTes, nTes, nTes)
        lines = lines + line
        line = "Uaz3_Tes%d = Ufoam_Tes%d*ALat_Tes%d/3 ! @userDefined W/k\n" % (nTes, nTes, nTes)
        lines = lines + line
        line = "UaTop_Tes%d = Ufoam_Tes%d*ATop_Tes%d ! @userDefined W/k\n" % (nTes, nTes, nTes)
        lines = lines + line

        return lines

    def getFirst12Par(self, nTes):

        lines = ""
        line = "Vol_Tes%d     ! 1: m3, volume of store\n" % nTes
        lines = lines + line
        line = "RhoWat_Tes%d  ! 2: kg/m3, density of storage media\n" % nTes
        lines = lines + line
        line = "CpWat_Tes%d   ! 3: kJ/kgK, specific heat of storage media\n" % nTes
        lines = lines + line
        line = "lamZ_Tes%d    ! 4: W/mK, effective vertical thermal conductivity of TES\n" % nTes
        lines = lines + line
        line = "Heigh_Tes%d   ! 5: m, storage height\n" % nTes
        lines = lines + line
        line = "TIni_Tes%d   ! 6: oC, initial temperature\n" % nTes
        lines = lines + line
        line = "nCvMax_Tes%d  ! 7: -, minimum relative plug height\n" % nTes
        lines = lines + line
        line = "nCvMin_Tes%d  ! 8: -, maximum relative plug height\n" % nTes
        lines = lines + line
        line = "maxTDiff_Tes%d  ! 9: K, maximum temperature difference between plugs\n" % nTes
        lines = lines + line
        line = "readMode_Tes%d  ! 10: 1: from table, 0: Tini and CapTot\n" % nTes
        lines = lines + line
        line = "Tref_Tes%d     ! 11: oC, reference temperature\n" % nTes
        lines = lines + line

        return lines

    def getFirst12ParVar(self, nTes):

        lines = "CONSTANTS 12\n"
        line = "Vol_Tes%d=1          ! 1: m3, volume of store\n" % nTes
        lines = lines + line
        line = "RhoWat_Tes%d=RhoWat  ! 2: kg/m3, density of storage media\n" % nTes
        lines = lines + line
        line = "CpWat_Tes%d=CpWat    ! 3: kJ/kgK, specific heat of storage media\n" % nTes
        lines = lines + line
        line = "lamZ_Tes%d=0.6       ! 4: W/mK, effective vertical thermal conductivity of TES\n" % nTes
        lines = lines + line
        line = "Heigh_Tes%d=1.       ! 5: m, storage height\n" % nTes
        lines = lines + line
        line = "Tini_Tes%d=60.       ! 6: oC, initial temperature\n" % nTes
        lines = lines + line
        line = "nCvMax_Tes%d=400     ! 7: -, minimum relative plug height\n" % nTes
        lines = lines + line
        line = "nCvMin_Tes%d=20      ! 8: -, maximum relative plug height\n" % nTes
        lines = lines + line
        line = "maxTDiff_Tes%d=0.015 ! 9: K, maximum temperature difference between plugs\n" % nTes
        lines = lines + line
        line = "readMode_Tes%d=0     ! 10: 1: from table, 0: Tini and CapTot\n" % nTes
        lines = lines + line
        line = "Tref_Tes%d=273.15    ! 11: oC, reference temperature\n" % nTes
        lines = lines + line
        line = "Tmax_Tes%d=100.       ! 6: oC, initial temperature\n" % nTes
        lines = lines + line

        line = self.getUaParValues(nTes, self.Ua)
        lines = lines + line

        return lines

    def getHeighDirectPortsPar(self, nPorts, nTes):

        lines = ""
        parId = 21
        for i in range(self.nMaxPorts):
            if i <= nPorts - 1:
                line = "zInDp%d_Tes%d zOutDp%d_Tes%d zero Dp%dStrat_Tes%d ! %d - %d: zIn, zOut, cp, strat\n" % (
                    i + 1,
                    nTes,
                    i + 1,
                    nTes,
                    i + 1,
                    nTes,
                    parId,
                    parId + 4,
                )
                lines = lines + line
            else:
                line = "-1 -1 zero zero ! %d - %d: zIn, zOut, cp, strat\n" % (parId, parId + 4)
                lines = lines + line

            parId = parId + 5

        return lines

    def getHeighSensorPar(self, nTes):

        lines = ""
        parId = 61
        for i in range(self.nMaxSensor):
            line = "zSen%d_Tes%d " % (i + 1, nTes)
            lines = lines + line

        line = "! %d-%d : relative storage temperature sensor heights \n" % (parId, parId + 10)
        lines = lines + line

        return lines

    def getHeatSourcesValues(self, nTes, nHeatSources, connectorAux):

        lines = ""
        line = self.sLine + "************ AUXILIAR HEATING**********\n"
        if nHeatSources > 0:
            line = "CONSTANTS %d\n" % (nHeatSources * 2)
            lines = lines + line

            for i in range(nHeatSources):
                line = "zAux%d_Tes%d=%.2f !connectDDck\n" % (i + 1, nTes, connectorAux[i]["zAux"])
                lines = lines + line
                line = "qAux%d_Tes%d=%.2f !connectDDck\n" % (i + 1, nTes, connectorAux[i]["qAux"])
                lines = lines + line

            line = "\n"
            lines = lines + line

        return lines

    def getPositionOfHeatSources(self, nTes, nHeatSources):

        lines = ""
        line = "** 20 height position for any heat source, e.g. electrical backup or heat pump condenser. Any position can be due to a different heat source\n"
        lines = lines + line
        for i in range(20):
            if i <= nHeatSources - 1:
                line = "zAux%d_Tes%d " % (i + 1, nTes)
                lines = lines + line
            else:
                line = "zero "
                lines = lines + line

        line = "\n"
        lines = lines + line

        return lines

    def getHeighAvgSensorParValues(self, nTes):

        lines = self.sLine + "** USER DEFINED AVERAGED TEMPERATURE SENSOR HEIGHTS. To be changed by user \n" + self.sLine

        line = "CONSTANTS 10\n"
        lines = lines + line
        z = 0.05
        for i in range(self.nMaxAvgSensor):
            line = "zSenAvgBot%d_Tes%d = %.2f\n" % (i + 1, nTes, z)
            lines = lines + line
            line = "zSenAvgTop%d_Tes%d = %.2f\n" % (i + 1, nTes, z + 0.1)
            lines = lines + line
            z = z + 0.2

        return lines

    def getHeighAvgSensorPar(self, nTes):

        lines = ""
        parId = 71
        for i in range(self.nMaxAvgSensor):
            line = (
                "zSenAvgBot%d_Tes%d zSenAvgTop%d_Tes%d ! %d-%d : relative position of lower and upper edge temeprature sensors\n"
                % (i + 1, nTes, i + 1, nTes, parId, parId + 1)
            )
            lines = lines + line
            parId = parId + 2

        return lines

    def getInsulationPlateParValues(self, nTes):

        lines = "************* MOVING PLATE *******************\n"

        lines = lines + "CONSTANTS 4\n"

        line = (
            "MoInsPlate_Tes%d=0 ! 0-2, Insulation Plate Mode: 0 = no insulation plate inside TES, 1 = insulation plate at fixed relative height, 2 = insulation plate at fixed temperature / density controlled \n"
            % nTes
        )
        lines = lines + line
        line = (
            "zInsPlate_Tes%d=0  ! 0-1, relative position of fixed height insulation plate inside TES (only for Insulation Plate Mode = 1\n"
            % nTes
        )
        lines = lines + line
        line = (
            "TinsPlate_Tes%d=0  ! oC, temperature at which moveable insulation plate floats in TES (only for Insulation Plate Mode = 2)\n"
            % nTes
        )
        lines = lines + line
        line = (
            "UAinsPlate_Tes%d=0 ! W/K, overall heat transfer coefficient across moveable insulation plate (including heat transfer in gap between plate and wall and in wall at the respective height)\n"
            % nTes
        )
        lines = lines + line

        return lines

    def getInsulationPlatePar(self, nTes):

        lines = ""
        line = (
            "MoInsPlate_Tes%d ! 81: 0-2, Insulation Plate Mode: 0 = no insulation plate inside TES, 1 = insulation plate at fixed relative height, 2 = insulation plate at fixed temperature / density controlled \n"
            % nTes
        )
        lines = lines + line
        line = (
            "zInsPlate_Tes%d  ! 82: 0-1, relative position of fixed height insulation plate inside TES (only for Insulation Plate Mode = 1\n"
            % nTes
        )
        lines = lines + line
        line = (
            "TinsPlate_Tes%d  ! 83: oC, temperature at which moveable insulation plate floats in TES (only for Insulation Plate Mode = 2)\n"
            % nTes
        )
        lines = lines + line
        line = (
            "UAinsPlate_Tes%d ! 84: W/K, overall heat transfer coefficient across moveable insulation plate (including heat transfer in gap between plate and wall and in wall at the respective height)\n"
            % nTes
        )
        lines = lines + line

        return lines

    def getInputs(self, inputs):

        nInputs = 69
        nHx = inputs["nHx"]
        nTes = inputs["nTes"]
        nPorts = inputs["nPorts"]
        nHeatSources = inputs["nHeatSources"]

        lines = ""
        line = "INPUTS %d\n" % nInputs
        lines = lines + line
        line = "************10 DIRECT PORTS INPUTS***************\n"
        lines = lines + line

        for idPort in range(self.nMaxPorts):  # This adds 30 inputs
            if idPort <= nPorts - 1:
                line = "Tdp%dIn_Tes%d Mfrdp%d_Tes%d Tdp%dInRev_Tes%d\n" % (
                    idPort + 1,
                    nTes,
                    idPort + 1,
                    nTes,
                    idPort + 1,
                    nTes,
                )

            else:
                line = "zero zero zero\n"
            lines = lines + line
        lines = lines + "****************\nTroomStore"  # This adds one input
        lines = lines + "\n***************** 6 HX INPUTS ******************\n"

        for idHx in range(self.nMaxHx):  # This adds 18 inputs
            if idHx <= nHx - 1:
                line = "Thx%dIn_Tes%d Mfrhx%d_Tes%d Thx%dInRev_Tes%d\n" % (
                    idHx + 1,
                    nTes,
                    idHx + 1,
                    nTes,
                    idHx + 1,
                    nTes,
                )
            else:
                line = "zero zero zero\n"
            lines = lines + line
        lines = lines + "***************** 20 HEAT SOURCE INPUTS ******************\n"

        for i in range(20):  # This adds 20 inputs
            if i <= nHeatSources - 1:
                line = "qAux%d_Tes%d " % (i + 1, nTes)
            else:
                line = "zero "
            lines = lines + line
        lines = lines + "\n"

        lines = lines + "****************** INTIAL INPUTS***********************\n"

        for i in range(69):
            line = "zero "
            lines = lines + line
            if i == 9 or i == 19 or i == 29 or i == 39 or i == 49 or i == 59 or i == 69:
                line = "\n"
                lines = lines + line

        line = "\n"
        lines = lines + line

        return lines

    def getOutputs(self, inputs):

        nUnit = inputs["nUnit"]
        nTes = inputs["nTes"]
        nHx = inputs["nHx"]
        nPorts = inputs["nPorts"]

        lines = "*****************OUTPUTS****************\n"

        nEq = nPorts
        line = "EQUATIONS %d\n" % nEq
        lines = lines + line

        for idPort in range(nPorts):
            line = "Qdp%d_Tes%d=[%d,%d] ! \n" % (idPort + 1, nTes, nUnit, 30 + idPort + 1)
            lines = lines + line

        nEq = 21
        line = "EQUATIONS %d\n" % nEq
        lines = lines + line

        line = "TAvg_Tes%d = [%d,180] ! Average storage temperature \n" % (nTes, nUnit)
        lines = lines + line
        line = "***Temperatures at 10 equallay distributed height \n"
        lines = lines + line
        counter = 21
        for i in range(10):
            height = 0.05 + 0.1 * i
            line = "T%d_Tes%d =[%d,%d] !temperature at %.2f \n" % (i + 1, nTes, nUnit, counter, height)
            lines = lines + line
            counter = counter + 1

        line = "***Temperatures at 10 sensors user defined height\n"
        lines = lines + line
        counter = 71
        for i in range(10):
            line = "Tsen%d_Tes%d =[%d,%d] ! temperature at user defined sensor height Tsen%d_Tes%d \n" % (
                i + 1,
                nTes,
                nUnit,
                counter,
                i + 1,
                nTes,
            )
            lines = lines + line
            counter = counter + 1

        nEq = nHx
        if nEq > 0:
            line = "EQUATIONS %d\n" % nEq
            lines = lines + line

        counter = 102
        for idHx in range(nHx):
            line = "Qhx%dOut_Tes%d=[%d,%d] ! \n" % (idHx + 1, nTes, nUnit, counter + 2)
            lines = lines + line
            counter = counter + 10

        line = "EQUATIONS 1\n"
        lines = lines + line

        line = "qHeatSource_Tes%d = [%d,181] ! Heat input of all auxiliary heat sources [kW]\n" % (nTes, nUnit)
        lines = lines + line

        line = "EQUATIONS 5\n"
        lines = lines + line

        line = "sumQv_Tes%d     = [%d,176] ! Heat input of all heat exchangers [kW]\n" % (nTes, nUnit)
        lines = lines + line
        line = "sumQLoss_Tes%d  = [%d,177] ! Heat Losses of the Tes [kW]\n" % (nTes, nUnit)
        lines = lines + line
        line = "sumQAcum_Tes%d  = [%d,178] ! Sensible accumulated heat [kW]\n" % (nTes, nUnit)
        lines = lines + line
        line = "sumQPorts_Tes%d = [%d,179] ! Heat Input by direct ports [kW]\n" % (nTes, nUnit)
        lines = lines + line
        line = "Imb_Tes%d       = [%d,64]  ! Heat Imbalance in Tes  IMB = sumQv - sumQLoss -sumQAcum + sumQPort\n" % (
            nTes,
            nUnit,
        )
        lines = lines + line

        return lines

    def getMonthyPrinter(self, nTes, nUnit, inputs):

        nPrinterUnit = nUnit + 1
        lines = ""
        line = "CONSTANTS 1 \n"
        lines = lines + line
        line = "unitPrinter_Tes%d = %d \n" % (nTes, nPrinterUnit)
        lines = lines + line
        line = "ASSIGN temp\TES%d_MO.Prt unitPrinter_Tes%d\n" % (nTes, nTes)
        lines = lines + line
        line = "UNIT %d TYPE 46\n" % nPrinterUnit
        lines = lines + line
        line = "PARAMETERS 5\n"
        lines = lines + line
        line = "unitPrinter_Tes%d ! 1: Logical unit number, -\n" % nTes
        lines = lines + line
        line = "-1  ! 2: Logical unit for monthly summaries\n"
        lines = lines + line
        line = "1 ! 3: Relative or absolute start time. 0: print at time intervals relative to the simulation start time. 1: print at absolute time intervals. No effect for monthly integrations\n"
        lines = lines + line
        line = "-1  ! 4: Printing & integrating interval, h. -1 for monthly integration\n"
        lines = lines + line
        line = "0  ! 5: Number of inputs to avoid integration\n"
        lines = lines + line
        # line = "0  ! 6: Number of inputs to avoid integration\n"; lines = lines + line
        nInputs = 5 + inputs["nPorts"] + inputs["nHx"] + 1  # +inputs["nHeatSources"]

        line = "INPUTS %d\n" % nInputs
        lines = lines + line
        line = "sumQv_Tes1 sumQLoss_Tes1 sumQAcum_Tes1 sumQPorts_Tes1 Imb_Tes1 "
        lines = lines + line
        for i in range(inputs["nPorts"]):
            line = "Qdp%d_Tes%d " % (i + 1, nTes)
            lines = lines + line

        for i in range(inputs["nHx"]):
            line = "Qhx%dOut_Tes%d " % (i + 1, nTes)
            lines = lines + line

        line = "qHeatSource_Tes%d " % (nTes)
        lines = lines + line

        lines = lines + "\n"

        for i in range(nInputs):
            line = "zero "
            lines = lines + line

        lines = lines + "\n"

        return lines

    def getOnlinePlotter(self, nTes, nUnit, inputs):

        nPrinterUnit = nUnit + 1
        lines = ""
        lines = lines + self.sLine
        line = "********** Online Plotter ***********\n"
        lines = lines + line
        lines = lines + self.sLine
        line = "\n"
        lines = lines + line
        line = "UNIT 501 TYPE 65     ! Online Plotter HX \n"
        lines = lines + line
        line = "PARAMETERS 12   \n"
        lines = lines + line
        line = "10     ! 1 Nb. of left-axis variables \n"
        lines = lines + line
        line = "0     ! 2 Nb. of right-axis variables\n"
        lines = lines + line
        line = "0     ! 3 Left axis minimum \n"
        lines = lines + line
        line = "100     ! 4 Left axis maximum -\n"
        lines = lines + line
        line = "0     ! 5 Right axis minimum \n"
        lines = lines + line
        line = "100     ! 6 Right axis maximum \n"
        lines = lines + line
        line = "nPlotsPerSim     ! 7 Number of plots per simulation \n"
        lines = lines + line
        line = "12     ! 8 X-axis gridpoints\n"
        lines = lines + line
        line = "1     ! 9 Shut off Online w/o removing \n"
        lines = lines + line
        line = "-1     ! 10 Logical unit for output file \n"
        lines = lines + line
        line = "0     ! 11 Output file units\n"
        lines = lines + line
        line = "0     ! 12 Output file delimiter\n"
        lines = lines + line
        line = "INPUTS 10     \n"
        lines = lines + line
        line = "T1_Tes%d T2_Tes%d T3_Tes%d T4_Tes%d T5_Tes%d T6_Tes%d T7_Tes%d T8_Tes%d T9_Tes%d T10_Tes%d \n" % (
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
        )
        lines = lines + line
        line = "T1_Tes%d T2_Tes%d T3_Tes%d T4_Tes%d T5_Tes%d T6_Tes%d T7_Tes%d T8_Tes%d T9_Tes%d T10_Tes%d \n" % (
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
            nTes,
        )
        lines = lines + line
        line = "LABELS  3         \n"
        lines = lines + line
        line = "Temperatures  \n"
        lines = lines + line
        line = "MassFlows  \n"
        lines = lines + line
        line = "Tes%d \n"
        lines = lines + line

        lines = lines + "\n"

        lines = lines + "\n"

        return lines

    def getParameters(self, inputs):

        nUnit = inputs["nUnit"]
        nType = inputs["nType"]
        nTes = inputs["nTes"]
        nPorts = inputs["nPorts"]
        nHx = inputs["nHx"]
        nHeatSources = inputs["nHeatSources"]

        lines = self.sLine + "********** TYPE DEFINITION **********\n" + self.sLine

        lines = lines + "UNIT %d TYPE %d     ! plug flow tank\n" % (nUnit, nType)
        lines = lines + "PARAMETERS 219 \n"
        lines = lines + self.getFirst12Par(nTes)
        lines = lines + self.getUaPar(nTes)
        lines = lines + "tMax_Tes%d\n" % nTes
        lines = lines + "0\t0\t0 ! 17-20 unsused parameters\n"
        lines = lines + self.getHeighDirectPortsPar(nPorts, nTes)
        lines = lines + self.getHeighSensorPar(nTes)
        lines = lines + self.getHeighAvgSensorPar(nTes)
        lines = lines + self.getInsulationPlatePar(nTes)
        lines = lines + "nHxUsed_Tes%d     ! 85: number Of used Hx\n" % nTes

        lines = lines + self.getHxParValues(nTes, nHx)
        lines = lines + self.getPositionOfHeatSources(nTes, nHeatSources)
        lines = lines + self.getInputs(inputs)
        lines = lines + self.getOutputs(inputs)
        lines = lines + self.getMonthyPrinter(nTes, nUnit, inputs)
        lines = lines + self.getOnlinePlotter(nTes, nUnit, inputs)
        return lines

    def getHead(self):

        header = open("C:\Daten\OngoingProject\SolTherm2050\Simulations\ddck\Generic\Head.ddck", "r")
        lines = header.read()
        header.close()

        return lines

    def createDDck(self, path, tankName, typeFile="ddck"):
        lines = ""
        if typeFile == "ddck":
            self.extension = "ddck"
            lines = lines + self.sLine
            lines = lines + ("**BEGIN %s.ddck\n" % tankName)
            lines = lines + self.sLine + "\n"
            lines = lines + self.sLine
            lines = lines + "** Plug-Flow Model exported from TRNSYS GUI\n"
            lines = lines + self.sLine + "\n"
            lines = lines + self.sLine
            lines = lines + "** To be checked: \n"
            lines = lines + "** check cp and rho values for the cirquits \n"
            lines = lines + "** default is cpwat and rhowat, for solarcirc usually cpbri and rhobri have to be used \n"
            lines = lines + self.sLine + "\n"
            lines = lines + self.sLine
            lines = lines + "** outputs to energy balance in kWh\n"
            lines = (
                lines + "** Following this naming standard : qSysIn_name, qSysOut_name, elSysIn_name, elSysOut_name\n"
            )
            lines = lines + self.sLine
            lines = lines + "EQUATIONS 3\n"
            lines = lines + ("qSysOut_Tes%sLoss = sumQLoss_Tes%d\n" % (tankName, self.inputs["nTes"]))
            lines = lines + ("qSysOut_Tes%sAcum = sumQAcum_Tes%d\n" % (tankName, self.inputs["nTes"]))
            lines = lines + ("elSysIn_Q_Tes%sAux = qHeatSource_Tes%d\n" % (tankName, self.inputs["nTes"]))

        elif typeFile == "dck":
            self.extension = "dck"
            lines = self.getHead()
        else:
            raise ValueError("typeFile %s unknown (Must be dck or ddck)")

        lines = lines + "\n" + self.sLine + "*** Inputs from hydraulic solver ****\n" + self.sLine

        nTes = self.inputs["nTes"]
        nHxs = self.inputs["nHx"]
        nPorts = self.inputs["nPorts"]
        nAux = self.inputs["nHeatSources"]
        nUnit = self.inputs["nUnit"]

        ddcxLines = ""

        for idPort in range(self.nMaxPorts):
            if idPort <= nPorts - 1:
                line = self._getOnePortInputs(nTes, idPort + 1, self.connectorsPort)
                lines = lines + line
                inputTemperature = line.split("\n")[1].split("=")[0].replace(" ", "")
                ddcxLine = (
                        "T"
                        + tankName
                        + "Port"
                        + self.connectorsPort[idPort]["side"]
                        + str(int(round(self.connectorsPort[idPort]["zIn"] * 100, 0)))
                        + "="
                        + inputTemperature
                        + "\n"
                )

                ddcxLines = ddcxLines + ddcxLine

        for idHx in range(self.nMaxHx):
            if idHx <= nHxs - 1:
                line = self.getOneHxInputs(nTes, idHx + 1, self.connectorsHx)
                lines = lines + line

        line = self.getHeatSourcesValues(nTes, nAux, self.connectorsAux)
        lines = lines + line

        lines = lines + self.sLine + "**** Outputs to hydraulic solver ****\n" + self.sLine

        nEq = nPorts
        if nEq > 0:
            line = "EQUATIONS %d\n" % nEq
            lines = lines + line
            line = "*** direct port outputs\n"
            lines = lines + line

        counter = 1
        for idPort in range(nPorts):
            outputTemperature = "Tdp%dOut_Tes%d" % (idPort + 1, nTes)
            line = outputTemperature + "=[%d,%d] ! \n" % (nUnit, counter)
            lines = lines + line
            ddcxLine = (
                    "T"
                    + tankName
                    + "Port"
                    + self.connectorsPort[idPort]["side"]
                    + str(int(self.connectorsPort[idPort]["zOut"] * 100))
                    + "="
                    + outputTemperature
                    + "\n"
            )
            ddcxLines = ddcxLines + ddcxLine
            counter = counter + 2

        nEq = nHxs
        if nEq > 0:
            line = "EQUATIONS %d\n" % nEq
            lines = lines + line
            line = "*** heat exchanger outputs\n"
            lines = lines + line

        counter = 102
        for idHx in range(nHxs):
            outputTemperature = "Thx%dOut_Tes%d" % (idHx + 1, nTes)
            line = outputTemperature + "=[%d,%d] ! \n" % (nUnit, counter)
            lines = lines + line
            ddcxLine = "T" + self.connectorsHx[idHx]["Name"] + "=" + outputTemperature + "\n"
            ddcxLines = ddcxLines + ddcxLine
            # line="Qhx%dOut_Tes%d=[%d,%d] ! \n"%(idHx+1,nTes,nUnit,counter+2);lines=lines+line
            counter = counter + 10

        if ddcxLines != "":
            header = self.sLine + "**BEGIN " + tankName + ".ddcx\n" + self.sLine
            header = header + "** This file is used to store the black box component outputs of " + tankName + "\n\n"
            ddcxLines = header + ddcxLines
            outfileDdcxPath = os.path.join(path, tankName)
            outfileDdcx = open(outfileDdcxPath + ".ddcx", "w")
            outfileDdcx.writelines(ddcxLines)
            outfileDdcx.close()

        lines = lines + "\n"

        line = self.sLine + "****** Parameters of Type1924 *******\n" + self.sLine
        lines = lines + line

        if nTes == 1:
            lines = lines + "CONSTANTS 3\n"
            line = "TRoomStore=15 ! @userDefined\n"
            lines = lines + line
            line = "VStoreRef = 0.763\n"
            lines = lines + line
        else:
            lines = lines + "CONSTANTS 1\n"
        line = "ratioTes%d = Vol_Tes%d / VStoreRef\n" % (nTes, nTes)
        lines = lines + line

        for idPort in range(self.nMaxPorts):
            if idPort <= nPorts - 1:

                line = self.getOnePortParConn(idPort, nTes, self.connectorsPort)
                lines = lines + line

        for idHx in range(self.nMaxHx):
            if idHx <= nHxs - 1:
                line = self.getOneHxParConnValues(nTes, idHx + 1, self.connectorsHx)
                lines = lines + line

        for idPort in range(self.nMaxPorts):
            if idPort <= nPorts - 1:

                line = self.getOnePortPar(idPort, nTes)
                lines = lines + line

        lines = lines + "********** HEAT EXCHANGER CONSTANTS*******\n"
        lines = lines + "CONSTANTS 1\n"
        lines = lines + "nHxUsed_Tes%d=%d \n" % (nTes, nHxs)

        for idHx in range(self.nMaxHx):
            if idHx <= nHxs - 1:
                line = self.getOneHxParValues(nTes, idHx + 1)
                lines = lines + line

        lines = lines + self.getSensorPositionValues(nTes)
        lines = lines + self.getHeighAvgSensorParValues(nTes)

        lines = lines + self.getFirst12ParVar(nTes)
        lines = lines + self.getInsulationPlateParValues(nTes)
        lines = lines + self.getParameters(self.inputs)

        nameWithPath = "%s\%s.%s" % (path, (tankName), self.extension)
        outfile = open(nameWithPath, "w")
        outfile.writelines(lines)
        outfile.close()



# pylint: skip-file
# type: ignore

import dataclasses as _dc
import pathlib as _pl
import typing as _tp
import uuid as _uuid

import appdirs as _ad

import trnsysGUI.serialization as _ser


@_dc.dataclass
class SettingsVersion0(_ser.UpgradableJsonSchemaMixinVersion0):
    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID('f7a2c37c-ce82-4a4a-9869-80f00083275b')

    trnsysBinaryDirPath: str


@_dc.dataclass
class Settings(_ser.UpgradableJsonSchemaMixin):
    @staticmethod
    def create(trnsysBinaryDirPath: _pl.Path) -> "Settings":
        return Settings(str(trnsysBinaryDirPath))

    _SETTINGS_FILE_NAME = "settings.json"

    trnsysBinaryPath: str

    @classmethod
    def tryLoadOrNone(cls) -> _tp.Optional["Settings"]:
        settingsFilePath = cls._getSettingsFilePath()

        if not settingsFilePath.exists():
            return None

        data = settingsFilePath.read_text()

        return Settings.from_json(data)

    @classmethod
    def load(cls) -> "Settings":
        settings = cls.tryLoadOrNone()

        if not settings:
            raise RuntimeError("No settings found.")

        return settings

    def save(self) -> None:
        settingsFilePath = self._getSettingsFilePath()

        settingsDirPath = settingsFilePath.parent
        if not settingsDirPath.exists():
            settingsDirPath.mkdir(parents=True)

        data = self.to_json(indent=4)
        settingsFilePath.write_text(data)

    @classmethod
    def _getSettingsFilePath(cls) -> _pl.Path:
        userConfigDirPath = _pl.Path(_ad.user_config_dir("pytrnsys-gui", "SPF OST"))
        settingsFilePath = userConfigDirPath / cls._SETTINGS_FILE_NAME
        return settingsFilePath

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[SettingsVersion0]:
        return SettingsVersion0

    @classmethod
    def upgrade(cls, superseded: SettingsVersion0) -> "Settings":
        trnsysBinaryPath = _pl.Path(superseded.trnsysBinaryDirPath) / "TRNExe.exe"
        return Settings.create(trnsysBinaryPath)

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID('e5ea1fbd-1be9-4415-b3e9-7f3a2a11d216')



# pylint: skip-file
# type: ignore

import random

from trnsysGUI.Connection import Connection


class PipeDataHandler(object):
    def __init__(self, diagramEditor):
        self.massflowFilePath = "massflowData/flows.txt"
        self.timeSteps = 365
        self.diagramEditor = diagramEditor

    def clearFile(self):
        f = open(self.massflowFilePath, "w")
        f.truncate(0)
        f.close()

    def generateData(self):
        self.clearFile()
        f = open(self.massflowFilePath, "w")
        for ts in range(self.timeSteps):
            for t in self.diagramEditor.trnsysObj:
                if isinstance(t, Connection):
                    if ts != 100:
                        a = random.random()
                    else:
                        a = 0

                    if a < 1 / 3:
                        a = -1e3
                    elif 1 / 3 <= a <= 2 / 3:
                        a = 0
                    else:
                        a = 1e3
                    f.write(str(a) + " " * (10 - len(str(a))))

            f.write("\n")

        f.close()



# pylint: skip-file
# type: ignore

from PyQt5.QtWidgets import QDialog, QLabel, QPushButton, QListWidget, QHBoxLayout, QVBoxLayout


class GroupChooserBlockDlg(QDialog):
    def __init__(self, block, parent=None):
        super(GroupChooserBlockDlg, self).__init__(parent)
        headerLabel = QLabel("Pick a group for this block:")
        self.block = block

        self.okButton = QPushButton("OK")
        self.cancelButton = QPushButton("Cancel")

        self.listw = QListWidget()

        buttonLayout = QHBoxLayout()
        buttonLayout.addStretch()
        buttonLayout.addWidget(self.okButton)
        buttonLayout.addWidget(self.cancelButton)

        vertLayout = QVBoxLayout()
        vertLayout.addWidget(headerLabel)
        vertLayout.addWidget(self.listw)
        vertLayout.addLayout(buttonLayout)

        self.setLayout(vertLayout)

        self.okButton.clicked.connect(self.acceptedEdit)
        self.cancelButton.clicked.connect(self.cancel)
        self.setWindowTitle("Change group")

        self.loadList()

        self.show()

    def loadList(self):
        for g in self.block.parent.parent().groupList:
            self.listw.addItem(g.displayName)

    def acceptedEdit(self):
        if len(self.listw.selectedItems()) > 0:
            self.block.setBlockToGroup(self.listw.selectedItems()[0].text())
            self.close()

    def cancel(self):
        # print("Canceling")
        self.close()

    # def removeHxL(self):
    #     for i in self.storage.heatExchangers:
    #         # Name is identified through index of comma
    #         for j in self.listWL.selectedItems():
    #             if i.displayName == j.text()[:j.text().find(",")]:
    #                 self.storage.heatExchangers.remove(i)
    #                 self.listWL.takeItem(self.listWL.row(self.listWL.selectedItems()[0]))
    #                 self.storage.parent.scene().removeItem(i.port1)
    #                 self.storage.parent.scene().removeItem(i.port2)
    #                 self.storage.parent.scene().removeItem(i)



import dataclasses as _dc
import logging as _log
import pathlib as _pl
import re as _re
import shutil as _sh
import typing as _tp

import PyQt5.QtWidgets as _qtw
import pytest as _pt

import trnsysGUI.StorageTank as _st
import trnsysGUI.diagram.Editor as _de

_DATA_DIR = _pl.Path(__file__).parent / "data"


@_dc.dataclass
class _Project:
    projectName: str
    testCasesDirName: str
    shallCopyFolderFromExamples: bool

    @staticmethod
    def createForTestProject(projectName: str) -> "_Project":
        return _Project(projectName, "tests", False)

    @staticmethod
    def createForExampleProject(projectName: str) -> "_Project":
        return _Project(projectName, "examples", True)

    @property
    def testId(self) -> str:
        return f"{self.projectName} [in {self.testCasesDirName}]"


def getProjects() -> _tp.Iterable[_Project]:
    yield _Project.createForExampleProject("TRIHP_dualSource")

    yield from getTestProjects()


def getTestProjects() -> _tp.Iterable[_Project]:
    testProjectTestCasesDir = _DATA_DIR / "tests"
    testProjectTestCaseDirPaths = [tc for tc in testProjectTestCasesDir.iterdir() if not tc.name == "README.txt"]
    for testProjectTestCaseDirPath in testProjectTestCaseDirPaths:
        projectName = testProjectTestCaseDirPath.name
        yield _Project.createForTestProject(projectName)


TEST_CASES = [_pt.param(p, id=p.testId) for p in getProjects()]


class TestEditor:
    @_pt.mark.parametrize("project", TEST_CASES)
    def testStorageAndHydraulicExports(self, project: _Project, request: _pt.FixtureRequest):
        helper = _Helper(project)
        helper.setup()

        # The following line is required otherwise QT will crash
        application = _qtw.QApplication([])

        def quitApplication():
            application.quit()

        request.addfinalizer(quitApplication)

        projectFolderPath = helper.actualProjectFolderPath

        self._exportHydraulic(projectFolderPath, _format="mfs")
        mfsDdckRelativePath = f"{project.projectName}_mfs.dck"
        helper.ensureFilesAreEqual(mfsDdckRelativePath, shallReplaceRandomizedFlowRates=True)

        self._exportHydraulic(projectFolderPath, _format="ddck")
        hydraulicDdckRelativePath = "ddck/hydraulic/hydraulic.ddck"
        helper.ensureFilesAreEqual(hydraulicDdckRelativePath, shallReplaceRandomizedFlowRates=False)

        storageTanks = self._exportStorageTank(projectFolderPath)
        for storageTank in storageTanks:
            ddckFileRelativePath = f"ddck/{storageTank.displayName}/{storageTank.displayName}.ddck"
            helper.ensureFilesAreEqual(ddckFileRelativePath, shallReplaceRandomizedFlowRates=False)

            ddcxFileRelativePath = f"ddck/{storageTank.displayName}/{storageTank.displayName}.ddcx"
            helper.ensureFilesAreEqual(ddcxFileRelativePath, shallReplaceRandomizedFlowRates=False)

    def _exportStorageTank(self, projectFolderPath):
        editor = self._createEditor(projectFolderPath)
        storageTanks = [
            o
            for o in editor.trnsysObj
            if isinstance(o, _st.StorageTank)  # type: ignore[attr-defined] # pylint: disable=no-member
        ]
        for storageTank in storageTanks:
            storageTank.exportDck()
        return storageTanks

    @classmethod
    def _exportHydraulic(cls, projectFolderPath, *, _format):
        editor = cls._createEditor(projectFolderPath)
        editor.exportHydraulics(exportTo=_format)

    @staticmethod
    def _createEditor(projectFolderPath):
        logger = _log.Logger("root")
        editor = _de.Editor(
            parent=None,
            projectFolder=str(projectFolderPath),
            jsonPath=None,
            loadValue="load",
            logger=logger,
        )
        return editor


class _Helper:
    def __init__(
        self,
        project: _Project,
    ):
        self._project = project

        testCasesFolderPath = _DATA_DIR / self._project.testCasesDirName

        # project name is used twice because the project file must always be contained in a folder
        # of the same name
        self.actualProjectFolderPath = testCasesFolderPath / self._project.projectName / self._project.projectName

        self._expectedProjectFolderPath = testCasesFolderPath / self._project.projectName / "expected"

    def setup(self):
        if self._project.shallCopyFolderFromExamples:
            self._copyExampleToTestInputFolder()

    def ensureFilesAreEqual(self, fileRelativePathAsString: str, shallReplaceRandomizedFlowRates: bool):
        fileRelativePath = _pl.Path(fileRelativePathAsString)
        actualFilePath = self.actualProjectFolderPath / fileRelativePath
        expectedFilePath = self._expectedProjectFolderPath / fileRelativePath

        actualContent = actualFilePath.read_text()
        expectedContent = expectedFilePath.read_text()

        if shallReplaceRandomizedFlowRates:
            actualContent = self._replaceRandomizedFlowRatesWithPlaceHolder(actualContent, placeholder="XXX")

        assert actualContent == expectedContent

    def _copyExampleToTestInputFolder(self):
        if self.actualProjectFolderPath.exists():
            _sh.rmtree(self.actualProjectFolderPath)

        pytrnsysGuiDir = _pl.Path(__file__).parents[3]
        exampleFolderPath = pytrnsysGuiDir / "data" / "examples" / self._project.projectName

        _sh.copytree(exampleFolderPath, self.actualProjectFolderPath)

    @classmethod
    def _replaceRandomizedFlowRatesWithPlaceHolder(cls, actualContent, placeholder):
        pattern = r"^(?P<variableName>Mfr[^ \t=]+)[ \t]+=[ \t]+[0-9\.]+"

        def replaceValueWithPlaceHolder(match: _tp.Match):
            return cls._processMatch(match, placeholder)

        actualContent = _re.sub(pattern, replaceValueWithPlaceHolder, actualContent, flags=_re.MULTILINE)

        return actualContent

    @staticmethod
    def _processMatch(match: _tp.Match, placeholder: str) -> str:
        matchedText = match[0]
        if matchedText == "MfrsupplyWater = 1000":
            return matchedText

        variableName = match["variableName"]

        return f"{variableName} = {placeholder}"



import dataclasses as _dc
import typing as _tp

import PyQt5.QtCore as _qtc
import PyQt5.QtWidgets as _qwt

NewHeight = _tp.Union[float, _tp.Literal["empty"]]


@_dc.dataclass
class NewRelativeHeights:
    input: NewHeight
    output: NewHeight


class ModifyRelativeHeightsDialog(_qwt.QDialog):
    """
    A dialog box lets the user choose the path and the name of folder for a new project
    """

    def __init__(self, relativeInputHeight: float, relativeOutputHeight: float, parent=None):
        super().__init__(parent)

        self.newRelativeHeights: _tp.Optional[NewRelativeHeights] = None

        inputLabel = _qwt.QLabel(f"New input height (was {relativeInputHeight * 100:.0f}%): ")
        self._inputLineEdit = _qwt.QLineEdit()
        inputLayout = _qwt.QHBoxLayout()
        inputLayout.addWidget(inputLabel)
        inputLayout.addWidget(self._inputLineEdit)

        outputLabel = _qwt.QLabel(f"New output height (was {relativeOutputHeight * 100:.0f}%): ")
        self._outputLineEdit = _qwt.QLineEdit()
        outputLayout = _qwt.QHBoxLayout()
        outputLayout.addWidget(outputLabel)
        outputLayout.addWidget(self._outputLineEdit)

        okButton = _qwt.QPushButton("Done")
        okButton.setFixedWidth(50)
        buttonLayout = _qwt.QHBoxLayout()
        buttonLayout.addWidget(okButton, alignment=_qtc.Qt.AlignCenter)

        overallLayout = _qwt.QVBoxLayout()
        overallLayout.addLayout(inputLayout)
        overallLayout.addLayout(outputLayout)
        overallLayout.addLayout(buttonLayout)
        self.setLayout(overallLayout)
        self.setFixedWidth(400)

        okButton.clicked.connect(self._onOk)
        self.setWindowTitle("Modify direct port couple")
        self.exec()

    def _onOk(self):
        inputText = self._inputLineEdit.text()
        newRelativeInputHeight = self._getNewRelativeHeight(inputText, "input")
        if not newRelativeInputHeight:
            return

        outputText = self._outputLineEdit.text()
        newRelativeOutputHeight = self._getNewRelativeHeight(outputText, "output")
        if not newRelativeOutputHeight:
            return

        self.newRelativeHeights = NewRelativeHeights(newRelativeInputHeight, newRelativeOutputHeight)

        self.close()

    @classmethod
    def _getNewRelativeHeight(cls, relativeHeightInPercentAsText: str, portName: str) -> _tp.Optional[NewHeight]:
        if not relativeHeightInPercentAsText:
            return "empty"

        try:
            relativeHeightInPercent = int(relativeHeightInPercentAsText)
        except ValueError:
            cls._showFormattingError(portName)
            return None

        if not 1 <= relativeHeightInPercent <= 99:
            cls._showFormattingError(portName)
            return None

        return relativeHeightInPercent / 100

    @staticmethod
    def _showFormattingError(portName: str) -> None:
        messageBox = _qwt.QMessageBox()
        messageBox.setText(
            f"Value for '{portName}' must be an integer between 1 and 99 inclusive."
        )
        messageBox.exec()



# pylint: skip-file
# type: ignore

import pathlib as _pl

from PyQt5 import QtCore
from PyQt5.QtGui import QCursor, QPixmap
from PyQt5.QtWidgets import QGraphicsPixmapItem, QMenu, QFileDialog

import trnsysGUI.imageAccessor as _ia
import trnsysGUI.images as _img
from trnsysGUI.ResizerItem import ResizerItem


class GraphicalItem(QGraphicsPixmapItem):
    def __init__(self, parent, **_):

        super(GraphicalItem, self).__init__(None)
        self.w = 100
        self.h = 100
        self.parent = parent
        self.resizeMode = False
        self.id = self.parent.parent().idGen.getID()

        self.flippedH = False
        self.flippedV = False
        self.rotationN = 0
        # Initial icon
        self._imageAccessor = _img.GEAR_SVG
        pixmap = _img.GEAR_SVG.pixmap(width=self.w, height=self.h)
        self.setPixmap(pixmap)

        self.setFlags(self.ItemIsSelectable | self.ItemIsMovable)
        self.setFlag(self.ItemSendsScenePositionChanges, True)
        self.setCursor(QCursor(QtCore.Qt.PointingHandCursor))

        self.parent.parent().graphicalObj.append(self)

    def setItemSize(self, w, h):
        self.w, self.h = w, h

    def contextMenuEvent(self, event):
        menu = QMenu()

        a1 = menu.addAction("Load different image")
        a1.triggered.connect(self.loadAction)

        a2 = menu.addAction("Delete this block")
        a2.triggered.connect(self.deleteBlock)

        menu.exec_(event.screenPos())

    def loadAction(self):
        fileName = QFileDialog.getOpenFileName(self.parent.parent(), "Load image", filter="*.png *.svg")[0]
        if fileName[-3:] == "png" or fileName[-3:] == "svg":
            self.setImageSource(fileName)
            self.updateImage()
        else:
            print("No image picked, name is " + fileName)

    def setImageSource(self, s: str):
        self._imageAccessor = _ia.ImageAccessor.createForFile(_pl.Path(s))

    def updateImage(self):
        pixmap = self._imageAccessor.pixmap(width=self.w, height=self.h)
        self.setPixmap(pixmap)

    def encode(self):
        dct = {}

        dct[".__BlockDict__"] = True
        dct["BlockName"] = "GraphicalItem"
        dct["BlockPosition"] = (float(self.pos().x()), float(self.pos().y()))
        dct["ID"] = self.id
        dct["Size"] = self.w, self.h
        dct["ImageSource"] = self._imageAccessor.getResourcePath()
        dct["FlippedH"] = self.flippedH
        dct["FlippedV"] = self.flippedV
        dct["RotationN"] = self.rotationN

        dictName = "GraphicalItem-"
        return dictName, dct

    def decode(self, i, resBlockList):
        self._imageAccessor = _ia.ImageAccessor.createFromResourcePath(i["ImageSource"])

        self.setPos(float(i["BlockPosition"][0]), float(i["BlockPosition"][1]))
        self.id = i["ID"]
        self.w, self.h = i["Size"]
        self.updateImage()

        resBlockList.append(self)

    def decodePaste(self, i, offset_x, offset_y, resConnList, resBlockList, **kwargs):
        self.setPos(float(i["BlockPosition"][0] + offset_x), float(i["BlockPosition"][1] + offset_y))

        # self.updateFlipStateH(i["FlippedH"])
        # self.updateFlipStateV(i["FlippedV"])
        # self.rotateBlockToN(i["RotationN"])

        resBlockList.append(self)

    def setParent(self, parent):
        self.parent = parent
        if self not in self.parent.parent().graphicalObj:
            self.parent.parent().graphicalObj.append(self)

    def deleteBlock(self):
        # self.parent.parent().trnsysObj.remove(self)
        self.parent.parent().graphicalObj.remove(self)
        self.parent.scene().removeItem(self)
        del self

    def mousePressEvent(self, event):
        try:
            self.resizer
        except AttributeError:
            self.resizer = ResizerItem(self)
            self.resizer.setPos(self.w, self.h)
            self.resizer.itemChange(self.resizer.ItemPositionChange, self.resizer.pos())
        else:
            return
        # self.resizer = ResizerItem(self)
        # self.resizer.setPos(self.w, self.h)
        # self.resizer.itemChange(self.resizer.ItemPositionChange, self.resizer.pos())

    def deleteResizer(self):
        del self.resizer



# pylint: skip-file
# type: ignore

import os
import shutil
import pathlib as _pl
import typing as _tp

from PyQt5.QtCore import QSize
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QMenu, QFileDialog, QTreeView

import trnsysGUI.images as _img
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.MyQFileSystemModel import MyQFileSystemModel
from trnsysGUI.MyQTreeView import MyQTreeView
from trnsysGUI.PortItem import PortItem


class GenericBlock(BlockItem):
    def __init__(self, trnsysType, parent, **kwargs):
        super(GenericBlock, self).__init__(trnsysType, parent, **kwargs)

        self.inputs.append(PortItem("i", 2, self))
        self.outputs.append(PortItem("o", 2, self))
        self.loadedFiles = []

        self.childIds = []
        self.childIds.append(self.trnsysId)

        self._imageAccessor = _img.GENERIC_BLOCK_PNG

        # Disallow adding port pairs later, because the trnsysIDs of the generated port pairs have to be
        # consecutive to be correctly printed out in the export
        self.isSet = True

        self.changeSize()
        self.addTree()

    def _getImageAccessor(self) -> _tp.Optional[_img.ImageAccessor]:
        return _img.GENERIC_BLOCK_PNG

    def changeSize(self):
        w = self.w
        h = self.h
        delta = 4

        """ Resize block function """

        # Limit the block size:
        if h < 20:
            h = 20
        if w < 40:
            w = 40
        # center label:
        rect = self.label.boundingRect()
        lw, lh = rect.width(), rect.height()
        lx = (w - lw) / 2
        self.label.setPos(lx, h)

        self.outputs[0].setPos(2 * delta + w, h - 2 * delta)
        self.inputs[0].setPos(2 * delta + w, 2 * delta)

        return w, h

    def getPairNb(self, side):
        res = 0
        for i in self.inputs:
            if i.side == side:
                res += 1

        self.logger.debug("there are " + str(res) + " pairs on the side " + str(side))
        return res

    def addPortDlg(self):
        self.parent.parent().showGenericPortPairDlg(self)

    def addPort(self, io, relH):
        self.logger.debug(io)
        self.logger.debug(relH)

    def setImage(self):
        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

    def changeImage(self):
        name = str(self.pickImage().resolve())
        if name[-3:] == "png" or name[-3:] == "svg":
            self.setImageSource(name)
            self.setImage()
        else:
            self.logger.debug("No image picked, name is " + name)

    def setImageSource(self, name):
        self._imageAccessor = _img.ImageAccessor.createForFile(_pl.Path(name))

    def pickImage(self):
        return _pl.Path(QFileDialog.getOpenFileName(self.parent.parent(), filter="*.png *.svg")[0])

    def contextMenuEvent(self, event):
        menu = QMenu()

        a1 = menu.addAction("Launch NotePad++")
        a1.triggered.connect(self.launchNotepadFile)

        rr = _img.ROTATE_TO_RIGHT_PNG.icon()
        a2 = menu.addAction(rr, "Rotate Block clockwise")
        a2.triggered.connect(self.rotateBlockCW)

        ll = _img.ROTATE_LEFT_PNG.icon()
        a3 = menu.addAction(ll, "Rotate Block counter-clockwise")
        a3.triggered.connect(self.rotateBlockCCW)

        a4 = menu.addAction("Reset Rotation")
        a4.triggered.connect(self.resetRotation)

        b1 = menu.addAction("Print Rotation")
        b1.triggered.connect(self.printRotation)

        c1 = menu.addAction("Delete this Block")
        c1.triggered.connect(self.deleteBlock)

        c3 = menu.addAction("Set image")
        c3.triggered.connect(self.changeImage)

        if not self.isSet:
            c4 = menu.addAction("Add port")
            c4.triggered.connect(self.addPortDlg)

        menu.exec_(event.screenPos())

    def encode(self):
        if self.isVisible():
            self.logger.debug("Encoding a Generic Block")

            portListInputs = []
            portListOutputs = []

            for p in self.inputs:
                portListInputs.append(p.id)
            for p in self.outputs:
                portListOutputs.append(p.id)

            dct = {}
            dct[".__BlockDict__"] = True
            dct["BlockName"] = self.name
            dct["BlockDisplayName"] = self.displayName
            dct["BlockPosition"] = (float(self.pos().x()), float(self.pos().y()))
            dct["ID"] = self.id
            dct["trnsysID"] = self.trnsysId
            dct["PortsIDIn"] = portListInputs
            dct["PortsIDOut"] = portListOutputs
            dct["FlippedH"] = self.flippedH
            dct["FlippedV"] = self.flippedV
            dct["RotationN"] = self.rotationN
            dct["GroupName"] = self.groupName
            dct["Imagesource"] = self._imageAccessor.getResourcePath()
            dct["PortPairsNb"] = [self.getPairNb(i) for i in range(4)]

            dictName = "Block-"
            return dictName, dct

    def decode(self, i, resConnList, resBlockList):
        self.logger.debug("Portpair is " + str(i["PortPairsNb"]))
        correcter = 0
        for j in range(4):
            if j == 2:
                correcter = -1
            for k in range(i["PortPairsNb"][j] + correcter):
                self.addPortPair(j)

        super(GenericBlock, self).decode(i, resConnList, resBlockList)
        self._imageAccessor = _img.ImageAccessor.createFromResourcePath(i["Imagesource"])
        self.setImage()

    def decodePaste(self, i, offset_x, offset_y, resConnList, resBlockList, **kwargs):
        correcter = 0
        for j in range(4):
            if j == 2:
                correcter = -1
            for k in range(i["PortPairsNb"][j] + correcter):
                self.addPortPair(j)

        super(GenericBlock, self).decodePaste(i, offset_x, offset_y, resConnList, resBlockList)
        self._imageAccessor = _img.ImageAccessor.createFromResourcePath(i["Imagesource"])
        self.setImage()

    def addPortPair(self, side):
        h = self.h
        w = self.w
        delta = 4
        self.logger.debug("side is " + str(side))
        self.inputs.append(PortItem("i", side, self))
        self.outputs.append(PortItem("o", side, self))
        # Allocate id
        self.childIds.append(self.parent.parent().idGen.getTrnsysID())

        portNb = [0, 0, 0, 0]
        for i in self.inputs:
            if i.side == 0:
                distBetweenPorts = (self.h - 4 * delta) / (2 * self.getPairNb(0) - 1)
                self.logger.debug("distance betw ports " + str(distBetweenPorts))
                i.setPos(-2 * delta, 2 * delta + distBetweenPorts * portNb[0])
                portNb[0] += 1

                self.outputs[self.inputs.index(i)].setPos(-2 * delta, 2 * delta + distBetweenPorts * portNb[0])
                portNb[0] += 1
            elif i.side == 1:
                distBetweenPorts = (self.w - 4 * delta) / (2 * self.getPairNb(1) - 1)
                i.setPos(2 * delta + distBetweenPorts * portNb[1], -2 * delta)
                portNb[1] += 1

                self.outputs[self.inputs.index(i)].setPos(2 * delta + distBetweenPorts * portNb[1], -2 * delta)
                portNb[1] += 1

            elif i.side == 2:
                self.logger.debug("side == 2")
                distBetweenPorts = (self.h - 4 * delta) / (2 * self.getPairNb(2) - 1)
                self.logger.debug("side 2 dist betw ports is " + str(distBetweenPorts))
                i.setPos(2 * delta + w, 2 * delta + distBetweenPorts * portNb[2])
                self.logger.debug(2 * delta + distBetweenPorts * portNb[2])
                portNb[2] += 1

                self.outputs[self.inputs.index(i)].setPos(2 * delta + w, 2 * delta + distBetweenPorts * portNb[2])
                self.logger.debug(2 * delta + distBetweenPorts * portNb[2])
                portNb[2] += 1

            else:
                distBetweenPorts = (self.w - 4 * delta) / (2 * self.getPairNb(3) - 1)
                self.logger.debug("distance betw ports " + str(distBetweenPorts))
                i.setPos(2 * delta + distBetweenPorts * portNb[3], 2 * delta + h)
                portNb[3] += 1

                self.outputs[self.inputs.index(i)].setPos(2 * delta + distBetweenPorts * portNb[3], 2 * delta + h)
                portNb[3] += 1

    def removePortPairOnSide(self, side):
        for i in self.inputs:
            if i.side == side:
                self.removePortPair(self.inputs.index(i))
                return

    def removePortPair(self, n):
        self.inputs.remove(self.inputs[n])
        self.outputs.remove(self.outputs[n])

    def updateFlipStateH(self, state):
        self.flippedH = bool(state)

        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

    def updateFlipStateV(self, state):
        self.flippedV = bool(state)

        pixmap = self._getPixmap()
        self.setPixmap(pixmap)

    def exportParametersFlowSolver(self, descConnLength):
        f = ""
        for i in range(len(self.inputs)):
            c = self.inputs[i].connectionList[0]
            temp = (
                str(c.trnsysId) + " " + str(self.outputs[i].connectionList[0].trnsysId) + " 0 0 "
            )
            temp += " " * (descConnLength - len(temp))

            # Generic block will have a 2n-liner exportConnString
            self.exportConnsString += temp + "\n"
            f += temp + "!" + str(self.childIds[i]) + " : " + self.displayName + "X" + str(i+1) + "\n"

        return f

    def exportInputsFlowSolver1(self):
        return "0,0 " * len(self.inputs), len(self.inputs)

    def exportInputsFlowSolver2(self):
        f = ""
        for i in range(len(self.inputs)):
            f += " " + str(self.exportInitialInput) + " "

        return f, len(self.inputs)

    def exportOutputsFlowSolver(self, prefix, abc, equationNumber, simulationUnit):
        tot = ""
        for j in range(len(self.inputs)):
            for i in range(0, 3):

                if i < 2:
                    temp = (
                        prefix
                        + self.displayName
                        + "-X"
                        + str(j)
                        + "_"
                        + abc[i]
                        + "=["
                        + str(simulationUnit)
                        + ","
                        + str(equationNumber)
                        + "]\n"
                    )
                    tot += temp
                    self.exportEquations.append(temp)
                    # nEqUsed += 1  # DC
                equationNumber += 1  # DC-ERROR it should count anyway

        return tot, equationNumber, 2 * len(self.inputs)

    def getSubBlockOffset(self, c):
        for i in range(len(self.inputs)):
            if (
                self.inputs[i] == c.toPort
                or self.inputs[i] == c.fromPort
                or self.outputs[i] == c.toPort
                or self.outputs[i] == c.fromPort
            ):
                return i

    def addTree(self):
        """
        When a blockitem is added to the main window.
        A file explorer for that item is added to the right of the main window by calling this method
        """
        self.logger.debug(self.parent.parent())
        pathName = self.displayName
        if self.parent.parent().projectPath == "":
            self.path = self.parent.parent().projectFolder
        else:
            self.path = self.parent.parent().projectPath

        self.path = os.path.join(self.path, "ddck")
        self.path = os.path.join(self.path, pathName)
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        self.model = MyQFileSystemModel()
        self.model.setRootPath(self.path)
        self.model.setName(self.displayName)
        self.tree = MyQTreeView(self.model, self)
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(self.path))
        self.tree.setObjectName("%sTree" % self.displayName)
        for i in range(1, self.model.columnCount() - 1):
            self.tree.hideColumn(i)
        self.tree.setMinimumHeight(200)
        self.tree.setSortingEnabled(True)
        self.parent.parent().splitter.addWidget(self.tree)

    def updateTreePath(self, path):
        """
        When the user chooses the project path for the file explorers, this method is called
        to update the root path.
        """
        pathName = self.displayName
        self.path = os.path.join(path, "ddck")
        self.path = os.path.join(self.path, pathName)
        if not os.path.exists(self.path):
            os.makedirs(self.path)
        self.model.setRootPath(self.path)
        self.tree.setRootIndex(self.model.index(self.path))

    def deleteBlock(self):
        """
        Overridden method to also delete folder
        """
        self.logger.debug("Block " + str(self) + " is deleting itself (" + self.displayName + ")")
        self.deleteConns()
        self.parent.parent().trnsysObj.remove(self)
        self.logger.debug("deleting block " + str(self) + self.displayName)
        self.parent.scene().removeItem(self)
        widgetToRemove = self.parent.parent().findChild(QTreeView, self.displayName + "Tree")
        shutil.rmtree(self.path)
        self.deleteLoadedFile()
        try:
            widgetToRemove.hide()
        except AttributeError:
            self.logger.debug("Widget doesnt exist!")
        else:
            self.logger.debug("Deleted widget")
        del self

    def setName(self, newName):
        """
        Overridden method to also change folder name
        """
        self.displayName = newName
        self.label.setPlainText(newName)
        self.model.setName(self.displayName)
        self.tree.setObjectName("%sTree" % self.displayName)
        self.logger.debug(os.path.dirname(self.path))
        destPath = os.path.join(os.path.split(self.path)[0], self.displayName)
        if os.path.exists(self.path):
            os.rename(self.path, destPath)
            self.path = destPath
            self.logger.debug(self.path)



# pylint: skip-file
# type: ignore

import sys

sys.path.append("C:\\Users\\parad\\OneDrive\\Desktop\\pytrnsys\\pytrnsys")
import pytrnsys.rsim.runParallelTrnsys as runTrnsys
import pytrnsys.psim.processParallelTrnsys as processTrnsys

import os


class ProcessMain:
    def processAction(self, logger, path):
        pathConfig = path
        configFile = "process.config"
        processTool = processTrnsys.ProcessParallelTrnsys()

        fullPath = os.path.join(path, configFile)
        logger.info("Starting ProcessParallelTrnsys with " + fullPath)

        try:
            processTool.readConfig(pathConfig, configFile)
        except ValueError as e:
            logger.error("EXCEPTION WHILE TRYING TO EXECUTE ProcessParallelTrnsys.readConfig")
            errorStatement = ""
            for words in e.args:
                errorStatement += str(words)
            return True, errorStatement
        except OSError as e:
            logger.error("EXCEPTION WHILE TRYING TO EXECUTE ProcessParallelTrnsys.readConfig")
            return True, str(e)
        except:
            logger.error("UNDEFINED EXCEPTION WHILE TRYING TO EXECUTE ProcessParallelTrnsys.readConfig")
            return True, ""

        try:
            processTool.process()
            logger.info("Successfully executed ProcessParallelTrnsys with " + fullPath)
            return False, ""
        except ValueError as e:
            logger.error("EXCEPTION WHILE TRYING TO EXECUTE ProcessParallelTrnsys.process")
            errorStatement = ""
            for words in e.args:
                errorStatement += str(words)
            return True, errorStatement
        except OSError as e:
            logger.error("EXCEPTION WHILE TRYING TO EXECUTE ProcessParallelTrnsys.process")
            return True, str(e)
        except:
            logger.error("UNDEFINED EXCEPTION WHILE TRYING TO EXECUTE ProcessParallelTrnsys.process")
            return True, ""



# pylint: skip-file
# type: ignore

from PyQt5.QtCore import Qt, QModelIndex
from PyQt5.QtWidgets import QFileSystemModel


class MyQFileSystemModel(QFileSystemModel):
    """
    Self defined QFileSystemModel in order to override headerData method
    """

    def setName(self, name):
        self.name = name

    def headerData(self, section, orientation, role):
        """
        To change the column name to correspond to object names
        """
        if section == 0 and role == Qt.DisplayRole:
            return self.name
        if section == 4 and role == Qt.DisplayRole:
            return "Priority"
        else:
            return super(QFileSystemModel, self).headerData(section, orientation, role)

    def columnCount(self, parent=QModelIndex()):
        # return super(MyQFileSystemModel, self).columnCount() + 1
        return 1

    # def data(self, index, role):
    #     if index.column() == self.columnCount() - 1:
    #         if role == Qt.DisplayRole:
    #             return "0"
    #         if role == Qt.TextAlignmentRole:
    #             return Qt.AlignHCenter
    #     return super(MyQFileSystemModel, self).data(index, role)

    # def setData(self, index, value, role):
    #     print(self.columnCount() - 1, index.column(), value)
    #     if index.column() == self.columnCount() - 1:
    #         if role == Qt.DisplayRole:
    #             return value
    #     return super(MyQFileSystemModel, self).setData(index, value, role)
